#
# This file is part of LiteDRAM.
#
# Copyright (c) 2020-2021 Antmicro <www.antmicro.com>
# SPDX-License-Identifier: BSD-2-Clause

import argparse

from migen import *

from litex.build.generic_platform import *
from litex.build.sim import SimPlatform
from litex.build.sim.config import SimConfig

from litex.soc.integration.common import *
from litex.soc.integration.soc_core import *
from litex.soc.integration.builder import *
from litex.soc.integration.soc import *
from litex.soc.cores.bitbang import *
from litex.soc.cores.cpu import CPUS

from litedram.gen import LiteDRAMCoreControl
from litedram.modules import EM6GA16L
from litedram.core.controller import ControllerSettings
from litedram.phy.rpc.simphy import SimulationPHY

# Platform -----------------------------------------------------------------------------------------

_io = [
    # clocks added later
    ("sys_rst", 0, Pins(1)),

    ("serial", 0,
        Subsignal("source_valid", Pins(1)),
        Subsignal("source_ready", Pins(1)),
        Subsignal("source_data",  Pins(8)),
        Subsignal("sink_valid",   Pins(1)),
        Subsignal("sink_ready",   Pins(1)),
        Subsignal("sink_data",    Pins(8)),
    ),

    # RPC pads
    ("rpcdram", 0,
        Subsignal("clk_p", Pins(1)),
        Subsignal("clk_n", Pins(1)),
        Subsignal("cs_n",  Pins(1)),
        Subsignal("dqs_p", Pins(1)),
        Subsignal("dqs_n", Pins(1)),
        Subsignal("stb",   Pins(1)),
        Subsignal("db",    Pins(16)),
    ),
]

class Platform(SimPlatform):
    def __init__(self):
        print('_io', end=' = '); __import__('pprint').pprint(_io)
        SimPlatform.__init__(self, "SIM", _io)

# DFI PHY model settings ---------------------------------------------------------------------------

sdram_module_nphases = {
    "SDR":   1,
    "DDR":   2,
    "LPDDR": 2,
    "DDR2":  2,
    "DDR3":  4,
    "RPC":   4,
    "DDR4":  4,
}

def get_sdram_phy_settings(memtype, data_width, clk_freq):
    nphases = sdram_module_nphases[memtype]
    assert memtype == "DDR3"

    # Settings from s7ddrphy
    tck                 = 2/(2*nphases*clk_freq)
    cmd_latency         = 0
    cl, cwl             = get_cl_cw(memtype, tck)
    cl_sys_latency      = get_sys_latency(nphases, cl)
    cwl                 = cwl + cmd_latency
    cwl_sys_latency     = get_sys_latency(nphases, cwl)
    rdcmdphase, rdphase = get_sys_phases(nphases, cl_sys_latency, cl)
    wrcmdphase, wrphase = get_sys_phases(nphases, cwl_sys_latency, cwl)
    read_latency        = 2 + cl_sys_latency + 2 + 3
    write_latency       = cwl_sys_latency

    sdram_phy_settings = {
        "nphases":       nphases,
        "rdphase":       rdphase,
        "wrphase":       wrphase,
        "rdcmdphase":    rdcmdphase,
        "wrcmdphase":    wrcmdphase,
        "cl":            cl,
        "cwl":           cwl,
        "read_latency":  read_latency,
        "write_latency": write_latency,
    }

    return PhySettings(
        phytype      = "SDRAMPHYModel",
        memtype      = memtype,
        databits     = data_width,
        dfi_databits = data_width if memtype == "SDR" else 2*data_width,
        **sdram_phy_settings,
    )

# Clocks -------------------------------------------------------------------------------------------

class Clocks(dict):
    # FORMAT: {name: {"freq_hz": _, "phase_deg": _}, ...}
    def names(self):
        return list(self.keys())

    def add_io(self, io):
        for name in self.names():
            print((name + "_clk", 0, Pins(1)))
            io.append((name + "_clk", 0, Pins(1)))

    def add_clockers(self, sim_config):
        for name, desc in self.items():
            sim_config.add_clocker(name + "_clk", **desc)

class _CRG(Module):
    def __init__(self, platform, domains=None):
        if domains is None:
            domains = ["sys"]
        # request() before clreating domains to avoid signal renaming problem
        domains = {name: platform.request(name + "_clk") for name in domains}

        self.clock_domains.cd_por = ClockDomain(reset_less=True)
        for name in domains.keys():
            setattr(self.clock_domains, "cd_" + name, ClockDomain(name=name))

        int_rst = Signal(reset=1)
        self.sync.por += int_rst.eq(0)
        self.comb += self.cd_por.clk.eq(self.cd_sys.clk)

        for name, clk in domains.items():
            cd = getattr(self, "cd_" + name)
            self.comb += cd.clk.eq(clk)
            self.comb += cd.rst.eq(int_rst)

# Simulation SoC -----------------------------------------------------------------------------------

class SimSoC(SoCCore):
    def __init__(self, clocks, trace_reset=1, auto_precharge=False, with_refresh=True, **kwargs):
        platform     = Platform()
        sys_clk_freq = clocks["sys"]["freq_hz"]

        # SoCCore ----------------------------------------------------------------------------------
        SoCCore.__init__(self, platform, clk_freq=sys_clk_freq,
            ident       = "LiteX Simulation",
            cpu_variant = "lite",
            **kwargs)

        # CRG --------------------------------------------------------------------------------------
        self.submodules.crg = _CRG(platform, clocks.names())

        # Debugging --------------------------------------------------------------------------------
        platform.add_debug(self, reset=trace_reset)

        # RPC DRAM ---------------------------------------------------------------------------------
        sdram_module = EM6GA16L(sys_clk_freq, "1:4")
        pads = platform.request("rpcdram")
        self.submodules.ddrphy = SimulationPHY(pads, sys_clk_freq=sys_clk_freq, generate_read_data=True)
        self.add_csr("ddrphy")

        controller_settings = ControllerSettings()
        controller_settings.auto_precharge = auto_precharge
        controller_settings.with_refresh = with_refresh

        self.add_sdram("sdram",
            phy                     = self.ddrphy,
            module                  = sdram_module,
            origin                  = self.mem_map["main_ram"],
            size                    = kwargs.get("max_sdram_size", 0x40000000),
            l2_cache_size           = kwargs.get("l2_size", 8192),
            l2_cache_min_data_width = kwargs.get("min_l2_data_width", 128),
            l2_cache_reverse        = False,
            controller_settings     = controller_settings
        )
        # Reduce memtest size for simulation speedup
        self.add_constant("MEMTEST_DATA_SIZE", 8*1024)
        self.add_constant("MEMTEST_ADDR_SIZE", 8*1024)

        self.add_constant("CONFIG_SIM_DISABLE_BIOS_PROMPT")
        # self.add_constant("CONFIG_DISABLE_DELAYS")

        self.submodules.ddrctrl = LiteDRAMCoreControl()
        self.add_csr("ddrctrl")
        self.sync += If(self.ddrctrl.init_done.storage, Finish())

        # Print info
        def dump(obj):
            print()
            print(" " + obj.__class__.__name__)
            print(" " + "-" * len(obj.__class__.__name__))
            d = obj if isinstance(obj, dict) else vars(obj)
            for var, val in d.items():
                if var == "self":
                    continue
                print("  {}: {}".format(var, val))
        print("=" * 80)
        dump(clocks)
        dump(self.ddrphy.settings)
        dump(sdram_module.geom_settings)
        dump(sdram_module.timing_settings)
        print()
        print("=" * 80)

# Build --------------------------------------------------------------------------------------------

def generate_gtkw_savefile(builder, vns, trace_fst):
    from litex.build.sim import gtkwave as gtkw

    dumpfile = os.path.join(builder.gateware_dir, "sim.{}".format("fst" if trace_fst else "vcd"))
    savefile = os.path.join(builder.gateware_dir, "sim.gtkw")
    soc = builder.soc
    rdphase = soc.sdram.controller.settings.phy.rdphase
    wrphase = soc.sdram.controller.settings.phy.wrphase

    with gtkw.GTKWSave(vns, savefile=savefile, dumpfile=dumpfile) as save:
        save.clocks()
        save.add(soc.bus.slaves["main_ram"], mappers=[gtkw.wishbone_sorter(), gtkw.wishbone_colorer()])
        save.fsm_states(soc)
        # all dfi signals
        save.add(soc.ddrphy.dfi, mappers=[gtkw.dfi_sorter(), gtkw.dfi_in_phase_colorer()])
        # each phase in separate group
        with save.gtkw.group("dfi phaseX", closed=True):
            for i, phase in enumerate(soc.ddrphy.dfi.phases):
                save.add(phase, group_name="dfi p{}".format(i), mappers=[
                    gtkw.dfi_sorter(phases=False),
                    gtkw.dfi_in_phase_colorer(),
                ])
        # only dfi command signals
        save.add(soc.ddrphy.dfi, group_name="dfi commands", mappers=[
            gtkw.regex_filter(gtkw.suffixes2re(["cas_n", "ras_n", "we_n"])),
            gtkw.dfi_sorter(),
            gtkw.dfi_per_phase_colorer(),
        ])
        # only dfi data signals
        save.add(soc.ddrphy.dfi, group_name="dfi wrdata", mappers=[
            gtkw.regex_filter(["wrdata$", f"p{wrphase}.*wrdata_en$"]),
            gtkw.dfi_sorter(),
            gtkw.dfi_per_phase_colorer(),
        ])
        save.add(soc.ddrphy.dfi, group_name="dfi wrdata_mask", mappers=[
            gtkw.regex_filter(gtkw.suffixes2re(["wrdata_mask"])),
            gtkw.dfi_sorter(),
            gtkw.dfi_per_phase_colorer(),
        ])
        save.add(soc.ddrphy.dfi, group_name="dfi rddata", mappers=[
            gtkw.regex_filter(gtkw.suffixes2re(["rddata", f"p{rdphase}.*rddata_valid"])),
            gtkw.dfi_sorter(),
            gtkw.dfi_per_phase_colorer(),
        ])
        # dram pads
        save.group([s for s in vars(soc.ddrphy.pads).values() if isinstance(s, Signal)],
            group_name = "pads",
            mappers = [
                gtkw.regex_filter(gtkw.suffixes2re(["dqs_n", "clk_n"]), negate=True),
                gtkw.regex_sorter(["clk", "cs", "stb", "db", "dqs"]),
                gtkw.regex_colorer({
                    "yellow": gtkw.suffixes2re(["cs"]),
                    "orange": gtkw.suffixes2re(["db", "dqs"]),
                    "red": gtkw.suffixes2re(["stb"]),
                }),
            ],
        )


def main():
    parser = argparse.ArgumentParser(description="Generic LiteX SoC Simulation")
    builder_args(parser)
    soc_core_args(parser)
    parser.add_argument("--threads",         default=1,           help="Set number of threads (default=1)")
    parser.add_argument("--rom-init",        default=None,        help="rom_init file")
    parser.add_argument("--sdram-init",      default=None,        help="SDRAM init file")
    parser.add_argument("--sdram-verbosity", default=0,           help="Set SDRAM checker verbosity")
    parser.add_argument("--trace",           action="store_true", help="Enable Tracing")
    parser.add_argument("--trace-fst",       action="store_true", help="Enable FST tracing (default=VCD)")
    parser.add_argument("--trace-start",     default=0,           help="Cycle to start tracing")
    parser.add_argument("--trace-end",       default=-1,          help="Cycle to end tracing")
    parser.add_argument("--trace-reset",     default=1,           help="Tracing state at start")
    parser.add_argument("--opt-level",       default="O3",        help="Compilation optimization level")
    parser.add_argument("--sys-clk-freq",    default="100e6",     help="Core clock frequency")
    parser.add_argument("--auto-precharge",  action="store_true", help="Use DRAM auto precharge")
    parser.add_argument("--no-refresh",      action="store_true", help="Disable DRAM refresher")
    parser.add_argument("--gtkw-savefile",   action="store_true", help="Generate GTKWSave savefile")
    args = parser.parse_args()

    soc_kwargs     = soc_core_argdict(args)
    builder_kwargs = builder_argdict(args)

    sys_clk_freq = int(float(args.sys_clk_freq))
    clocks = Clocks({
        "sys":           dict(freq_hz=sys_clk_freq),
        "sys2x":         dict(freq_hz=2*sys_clk_freq),
        "sys4x":         dict(freq_hz=4*sys_clk_freq),
        "sys4x_90":      dict(freq_hz=4*sys_clk_freq, phase_deg=90),
        "sys4x_180":     dict(freq_hz=4*sys_clk_freq, phase_deg=180),
        "sys4x_90_ddr":  dict(freq_hz=2*4*sys_clk_freq, phase_deg=2*90),
        "sys4x_180_ddr": dict(freq_hz=2*4*sys_clk_freq, phase_deg=(2*180)%360),
    })

    clocks.add_io(_io)

    sim_config = SimConfig()
    clocks.add_clockers(sim_config)

    # Configuration --------------------------------------------------------------------------------

    cpu = CPUS[soc_kwargs.get("cpu_type", "vexriscv")]
    if soc_kwargs["uart_name"] == "serial":
        soc_kwargs["uart_name"] = "sim"
        sim_config.add_module("serial2console", "serial")
    if args.rom_init:
        soc_kwargs["integrated_rom_init"] = get_mem_data(args.rom_init, endianness=cpu.endianness)
    args.with_sdram = True
    soc_kwargs["integrated_main_ram_size"] = 0x0
    soc_kwargs["sdram_verbosity"]          = int(args.sdram_verbosity)

    # SoC ------------------------------------------------------------------------------------------
    soc = SimSoC(
        clocks         = clocks,
        trace_reset    = args.trace_reset,
        auto_precharge = args.auto_precharge,
        with_refresh   = not args.no_refresh,
        sdram_init     = [] if args.sdram_init is None else get_mem_data(args.sdram_init, endianness=cpu.endianness),
        l2_size        = args.l2_size,
        **soc_kwargs)

    # Build/Run ------------------------------------------------------------------------------------
    builder_kwargs["csr_csv"] = "csr.csv"
    builder = Builder(soc, **builder_kwargs)
    vns = builder.build(run=False, threads=args.threads, sim_config=sim_config,
        opt_level   = args.opt_level,
        trace       = args.trace,
        trace_fst   = args.trace_fst,
        trace_start = int(args.trace_start),
        trace_end   = int(args.trace_end))

    if args.gtkw_savefile:
        generate_gtkw_savefile(builder, vns, trace_fst=args.trace_fst)

    builder.build(build=False, threads=args.threads, sim_config=sim_config,
        opt_level   = args.opt_level,
        trace       = args.trace,
        trace_fst   = args.trace,
        trace_start = int(args.trace_start),
        trace_end   = int(args.trace_end)
    )

if __name__ == "__main__":
    main()



#
# This file is part of LiteDRAM.
#
# Copyright (c) 2021 Antmicro <www.antmicro.com>
# SPDX-License-Identifier: BSD-2-Clause

import copy
import collections
from operator import or_, and_
from functools import reduce
from typing import Tuple

from migen import *

from litex.soc.interconnect import stream
from litex.soc.interconnect.csr import AutoCSR, CSRStorage, CSR

from litedram.common import BitSlip, get_sys_latency, get_sys_phase, PhySettings, TappedDelayLine
from litedram.phy.dfi import Interface as DFIInterface, DFIRateConverter
from litedram.phy.utils import CommandsPipeline, bitpattern, delayed, HoldValid
from litedram.phy.lpddr5.commands import DFIPhaseAdapter, WCKSyncType


class LPDDR5Output:
    """Unserialized output of LPDDR5PHY. Has to be serialized by concrete implementation."""
    def __init__(self, databits, wck_ck_ratio):
        assert databits % 8 == 0
        self.reset_n = Signal()
        self.ck      = Signal(2)
        self.cs      = Signal()  # CK SDR
        self.ca      = [Signal(2) for _ in range(7)]  # CK DDR
        # WCK DDR
        self.dq_o    = [Signal(2*wck_ck_ratio) for _ in range(databits)]
        self.dq_i    = [Signal(2*wck_ck_ratio) for _ in range(databits)]
        self.dq_oe   = Signal()
        self.wck     = [Signal(2*wck_ck_ratio)   for _ in range(databits//8)]
        self.rdqs_o  = [Signal(2*wck_ck_ratio)   for _ in range(databits//8)]
        self.rdqs_i  = [Signal(2*wck_ck_ratio)   for _ in range(databits//8)]
        self.rdqs_oe = Signal()
        self.dmi_o   = [Signal(2*wck_ck_ratio) for _ in range(databits//8)]
        self.dmi_i   = [Signal(2*wck_ck_ratio) for _ in range(databits//8)]
        self.dmi_oe  = Signal()


def namedtuple(cls):
    # This is a simple workaround the lack of dataclasses module in Python <3.7
    # It will instead take given class and generate a new one that inherits
    # from namedtuple with fields that are retrieved from class annotations.
    assert cls.__base__ is object, "Supports only simple classes with no inheritance"
    # retrieve field names and generate namedtuple class
    fields = [name for name, _type in cls.__annotations__.items()]
    ntuple_cls = collections.namedtuple(cls.__name__, fields)
    # create a new type inheriting from the generated namedtuple
    new_cls = type(cls.__name__, (ntuple_cls, object), cls.__dict__.copy())
    return new_cls


@namedtuple
class FreqRange:
    mr:                 int                  # MR2[3:0] value
    data_rate:          Tuple[int, int]      # (> Mbps, <= Mbps)
    wl:                 Tuple[int, int]      # (Set A, Set B)
    t_wckenl_wr:        Tuple[int, int]      # (Set A, Set B)
    t_wckpre_static:    int
    t_wckpre_toggle_wr: int
    rl:                 Tuple[int, int, int] # (Set 0, Set 1, Set 2)
    t_wckenl_rd:        Tuple[int, int, int] # (Set 0, Set 1, Set 2)
    t_wckpre_toggle_rd: int
    n_rbtp:             int
    n_wr:               int
    n_wr_op:            int                  # MR2[7:4]

    @property
    def ck_freq(self):
        low, high = self.data_rate
        return (round(low / 4), round(high / 4))

    def for_set(self, wl_set, rl_set):
        wl_set = {"A": 0, "B": 1}[wl_set]
        return self._replace(
            wl          = self.wl[wl_set],
            t_wckenl_wr = self.t_wckenl_wr[wl_set],
            rl          = self.rl[rl_set],
            t_wckenl_rd = self.t_wckenl_rd[rl_set],
        )

# Taken from Tables 182, 183, 201 of JEDEC specification for LPDDR5
# WCK:CK 2:1 or 4:1, DVFSC diabled, Read Link ECC off
FREQUENCY_RANGES = {
    2: [
        #         MR       DR            WL                     RL                              nWR
        FreqRange(0b0000, (40,   533),  (4,  4),  (1, 1), 1, 3, ( 6,  6,  6), (0, 0, 0),  6, 0,  5, 0b0000),
        FreqRange(0b0001, (533,  1067), (4,  6),  (0, 2), 2, 3, ( 8,  8,  8), (0, 0, 0),  7, 0, 10, 0b0001),
        FreqRange(0b0010, (1067, 1600), (6,  8),  (1, 3), 2, 4, (10, 10, 12), (1, 1, 3),  8, 0, 14, 0b0010),
        FreqRange(0b0011, (1600, 2133), (8,  10), (2, 4), 3, 4, (12, 14, 14), (2, 4, 4),  8, 0, 19, 0b0011),
        FreqRange(0b0100, (2133, 2750), (8,  14), (1, 7), 4, 4, (16, 16, 18), (3, 3, 5), 10, 2, 24, 0b0100),
        FreqRange(0b0101, (2750, 3200), (10, 16), (3, 9), 4, 4, (18, 20, 20), (5, 7, 7), 10, 2, 28, 0b0101),
    ],
    4: [
        #         MR       DR            WL                      RL                              nWR
        FreqRange(0b0000, (40,   533),  (2,  2),  (0, 0),  1, 2, ( 3,  3,  3), (0, 0, 0),  3, 0,  3, 0b0000),
        FreqRange(0b0001, (533,  1067), (2,  3),  (0, 1),  1, 2, ( 4,  4,  4), (0, 0, 0),  4, 0,  5, 0b0001),
        FreqRange(0b0010, (1067, 1600), (3,  4),  (1, 2),  1, 2, ( 5,  5,  6), (1, 1, 2),  4, 0,  7, 0b0010),
        FreqRange(0b0011, (1600, 2133), (4,  5),  (1, 2),  2, 2, ( 6,  7,  7), (1, 2, 2),  4, 0, 10, 0b0011),
        FreqRange(0b0100, (2133, 2750), (4,  7),  (1, 4),  2, 2, ( 8,  8,  9), (2, 2, 3),  5, 1, 12, 0b0100),
        FreqRange(0b0101, (2750, 3200), (5,  8),  (2, 5),  2, 2, ( 9, 10, 10), (3, 4, 4),  5, 1, 14, 0b0101),
        FreqRange(0b0110, (3200, 3733), (6,  9),  (2, 5),  3, 2, (10, 11, 12), (3, 4, 5),  5, 2, 16, 0b0110),
        FreqRange(0b0111, (3733, 4267), (6,  11), (2, 7),  3, 2, (12, 13, 14), (4, 5, 6),  6, 2, 19, 0b0111),
        FreqRange(0b1000, (4267, 4800), (7,  12), (3, 8),  3, 2, (13, 14, 15), (5, 6, 7),  6, 3, 21, 0b1000),
        FreqRange(0b1001, (4800, 5500), (8,  14), (3, 9),  4, 2, (15, 16, 17), (6, 7, 8),  6, 4, 24, 0b1001),
        FreqRange(0b1010, (5500, 6000), (9,  15), (4, 10), 4, 2, (16, 17, 19), (6, 7, 9),  7, 4, 26, 0b1010),
        FreqRange(0b1011, (6000, 6400), (9,  16), (4, 11), 4, 2, (17, 18, 20), (7, 8, 10), 7, 4, 28, 0b1011),
    ]
}

def get_frange(twck, wck_ck_ratio):
    data_rate = 2 * 1/twck
    for frange in FREQUENCY_RANGES[wck_ck_ratio]:
        dr_min, dr_max = frange.data_rate
        if dr_min < data_rate/1e6 <= dr_max:
            return frange
    raise ValueError

class LPDDR5PHY(Module, AutoCSR):
    """Core logic of LPDDR5 PHY

    Implements LPDDR5 logic translating DFI commands into sequences on LPDDR5 pads, without handling
    data (de-)serialization. To create a PHY for specific devices, derive from this class and add
    serdes blocks between the signals in `self.out` and device pads.

    DFI commands
    ------------
    We don't currently use official DFI specification regarding CA commands being generated by MC
    and sent on DFI.address. Instead, regular cas/ras/we commands are translated to CA values by the
    PHY itself. Some commands use special encoding, refer to `DFIPhaseAdapter` class documentation
    for more information.

    Parameters
    ----------
    pads : object
        Object containing LPDDR5 pads.
    ck_freq : float
        Frequency of commands clock (CK).
    ser_latency : Latency
        Additional latency introduced due to signal serialization.
    des_latency : Latency
        Additional latency introduced during signal deserialization.
    phytype : str
        Name of the PHY (concrete implementation).
    cmd_delay : int
        Used to force cmd delay during initialization in BIOS.
    masked_write : bool
        Use MASKED-WRITE commands if True else use WRITE commands (data masking will not work).
    wck_ck_ratio : 2 or 4
        Specifies the WCK:CK ratio used.
    """
    def __init__(self, pads, *, ck_freq, phytype, ser_latency, des_latency, cmd_delay=None,
            masked_write=True, wck_ck_ratio=2, csr_cdc=None):
        self.pads        = pads
        self.memtype     = memtype     = "LPDDR5"
        self.nranks      = nranks      = 1 if not hasattr(pads, "cs_n") else len(pads.cs_n)
        self.databits    = databits    = len(pads.dq)
        self.addressbits = addressbits = 18  # for activate row address
        self.bankbits    = bankbits    = 7  # 4, but 7 bits needed for Mode Register address
        self.nphases     = nphases     = 1
        self.twck        = twck        = 1 / (wck_ck_ratio * ck_freq)
        self.ser_latency = ser_latency
        self.des_latency = des_latency
        assert databits % 8 == 0

        # Parameters -------------------------------------------------------------------------------
        frange = get_frange(twck, wck_ck_ratio).for_set(wl_set="A", rl_set=0)

        # Burst spans several CK cycles
        burst_len = 16
        burst_ck_cycles = burst_len // (2*wck_ck_ratio)

        # Bitslip introduces latency from 1 up to `cycles + 1` (sys)
        bitslip_cycles  = 1
        bitslip_range   = 1
        # Commands are sent over 2 CK and we count cl/cwl from the 2nd CK
        cmd_latency     = 1

        cl, cwl = frange.rl, frange.wl  # measured with respect to CK

        # Read latency
        # DFI cmd -> cmd buf -> PHY serializers -> DRAM -> Read Latency -> DQ data
        # -> PHY deserializers -> Bitslip -> Burst cycles -> StrideConverter -> DFI rddata
        read_data_delay = cmd_latency + ser_latency.sys + cl  # DFI cmd -> read data on DQ
        read_des_delay  = des_latency.sys + bitslip_cycles+bitslip_range + burst_ck_cycles  # DQ -> DFI rddata
        read_latency    = read_data_delay + read_des_delay

        # Write latency
        write_latency = cwl + cmd_latency

        # Registers --------------------------------------------------------------------------------
        self._rst = CSRStorage()

        self._wlevel_en     = CSRStorage()
        self._wlevel_strobe = CSR()

        self._dly_sel = CSRStorage(databits//8)

        self._rdly_dq_bitslip_rst = CSR()
        self._rdly_dq_bitslip     = CSR()

        self._wdly_dq_bitslip_rst = CSR()
        self._wdly_dq_bitslip     = CSR()

        # Add CDC in case of memory controller operating in different clock domain than this PHY.
        csr_cdc = csr_cdc or (lambda i: i)
        wlevel_strobe       = csr_cdc(self._wlevel_strobe.re)
        rdly_dq_bitslip_rst = csr_cdc(self._rdly_dq_bitslip_rst.re)
        rdly_dq_bitslip     = csr_cdc(self._rdly_dq_bitslip.re)
        wdly_dq_bitslip_rst = csr_cdc(self._wdly_dq_bitslip_rst.re)
        wdly_dq_bitslip     = csr_cdc(self._wdly_dq_bitslip.re)

        # PHY settings -----------------------------------------------------------------------------
        self.settings = PhySettings(
            phytype       = phytype,
            memtype       = memtype,
            databits      = databits,
            dfi_databits  = burst_len * databits,
            nranks        = nranks,
            nphases       = nphases,
            rdphase       = 0,
            wrphase       = 0,
            cl            = cl,
            cwl           = cwl,
            read_latency  = read_latency,
            write_latency = write_latency,
            cmd_latency   = cmd_latency,
            cmd_delay     = cmd_delay,
            bitslips      = 8,
        )
        self.settings.wck_ck_ratio  = wck_ck_ratio

        # DFI Interface ----------------------------------------------------------------------------
        self.dfi = dfi = DFIInterface(
            addressbits, bankbits, nranks, self.settings.dfi_databits, nphases)

        # # #

        self.submodules.adapter = DFIPhaseAdapter(dfi.p0, masked_write=masked_write)
        self.out = LPDDR5Output(databits, wck_ck_ratio)

        # CK ---------------------------------------------------------------------------------------
        self.comb += self.out.ck.eq(bitpattern("-_-_-_-_"))

        # Commands ---------------------------------------------------------------------------------
        # Commands are sent with SDR CS and DDR CA[6:0] clocked by CK. DFI command can translate to
        # 1 or 2 LPDDR5 commands. If we need two commands we send them in this and the next cycle,
        # if there is 1 command, we delay it 1 cycle to be consistent with all timing calculations.
        # It is illegal to send 2 DFI commands in 2 following cycles (second will be ignored). This
        # will in practice be limited by timing constraints.
        cmd_buf = stream.PipeValid([("cs", 1), ("ca_p", 7), ("ca_n", 7)])
        self.submodules += cmd_buf
        self.comb += [
            # ignore command if it comes just after the first one
            cmd_buf.sink.valid.eq(self.adapter.valid & ~cmd_buf.source.valid),
            # buffer the the second command
            cmd_buf.sink.cs.eq(self.adapter.cmd2.cs),
            cmd_buf.sink.ca_p.eq(self.adapter.cmd2.ca[0]),
            cmd_buf.sink.ca_n.eq(self.adapter.cmd2.ca[1]),
            # we pop the data in the next cycle
            cmd_buf.source.ready.eq(1),
        ]

        def set_cmd(out, cmd1, cmd2, default=0):
            return If(cmd_buf.source.valid, # cmd2 stored in the previous cycle
                out.eq(cmd2)
            ).Elif(self.adapter.valid, # cmd1 on DFI (note: there is no cmd2 from prev cycle)
                out.eq(cmd1)
            ).Else(
                out.eq(default)
            )

        self.comb += set_cmd(self.out.cs, cmd1=self.adapter.cmd1.cs, cmd2=cmd_buf.source.cs),
        for bit in range(7):
            self.comb += [
                set_cmd(self.out.ca[bit][0], cmd1=self.adapter.cmd1.ca[0][bit], cmd2=cmd_buf.source.ca_p[bit]),
                set_cmd(self.out.ca[bit][1], cmd1=self.adapter.cmd1.ca[1][bit], cmd2=cmd_buf.source.ca_n[bit]),
            ]

        self.comb += self.out.reset_n.eq(self.dfi.p0.reset_n)

        # WCK --------------------------------------------------------------------------------------
        # WCK can be enabled/disabled. When enabling, it has to be synchronized with CK. To do so,
        # CAS (alone or followed by WR/RD) must be issued. Synchronization is done after tCKSENL_x
        # after CAS command (CK rising edge), by keeping WCK static for tWCKPRE_Static, then
        # toggling it for tWCKPRE_Toggle_x. If using WCK:CK=4:1, then the first CK of toggling
        # should be with half WCK frequency.
        # Timings are in relation to WL and RL as:
        # WL = tWCKENL_WR - 1 + tWCKPRE_Static + tWCKPRE_Toggle_WR
        # RL = tWCKENL_RD - 1 + tWCKPRE_Static + tWCKPRE_Toggle_RD  (without Byte Mode, nor Read DBI/Read Data Copy)
        wck_sync_done = Signal()
        wck_sync = TappedDelayLine(
            signal = self.adapter.wck_sync,
            ntaps  = max(1, max(frange.t_wckenl_wr, frange.t_wckenl_rd) + frange.t_wckpre_static),
        )
        self.submodules += wck_sync
        wck_sync_taps = Array([wck_sync.input, *wck_sync.taps])

        wck_pattern = Signal(8)  # for WCK:CK=2:1 we take wck_pattern[::2]
        patterns = {
            "disabled":      "________",
            "static":        "________",
            "toggle":        "--__--__",
            "toggle_4:1":    "-_-_-_-_",
            "postamble":     "--______",
            "postamble_4:1": "-_-_-___",
        }

        assert frange.t_wckpre_static > 0  # The algorithm assumes it's never 0

        wck_fsm = FSM()
        self.submodules += wck_fsm
        wck_fsm.act("DISABLED",
            wck_pattern.eq(bitpattern(patterns["disabled"])),
            If(wck_sync_taps[frange.t_wckenl_wr] == WCKSyncType.WR,
                NextState("STATIC")
            ).Elif(wck_sync_taps[frange.t_wckenl_rd] == WCKSyncType.RD,
                NextState("STATIC")
            )
        )
        wck_fsm.act("STATIC",
            wck_pattern.eq(bitpattern(patterns["static"])),
            If(wck_sync_taps[frange.t_wckenl_wr + frange.t_wckpre_static] == WCKSyncType.WR,
                NextState("TOGGLE")
            ).Elif(wck_sync_taps[frange.t_wckenl_rd + frange.t_wckpre_static] == WCKSyncType.RD,
                NextState("TOGGLE")
            )
        )
        wck_fsm.act("TOGGLE",
            wck_pattern.eq(bitpattern(patterns["toggle"])),
            If(~wck_sync_done,
                NextState("POSTAMBLE")
            ).Elif((wck_ck_ratio == 4),  # go to full speed in the next cycle
                NextState("TOGGLE_4:1")
            ),
        )
        wck_fsm.act("TOGGLE_4:1",
            wck_pattern.eq(bitpattern(patterns["toggle_4:1"])),
            If(~wck_sync_done,
                NextState("POSTAMBLE")
            ),
        )
        wck_fsm.act("POSTAMBLE",
            If((wck_ck_ratio == 4),
                wck_pattern.eq(bitpattern(patterns["postamble_4:1"])),
                NextState("DISABLED")
            ).Else(
                wck_pattern.eq(bitpattern(patterns["toggle"])),
                NextState("POSTAMBLE_2:1")
            )
        )
        wck_fsm.act("POSTAMBLE_2:1",
            wck_pattern.eq(bitpattern(patterns["postamble"])),
            NextState("DISABLED")
        )

        wck_out = {2: wck_pattern[::2], 4: wck_pattern}[wck_ck_ratio]
        assert len(wck_out) == len(self.out.wck[0]), (len(wck_out), len(self.out.wck))

        # WCK2CK leveling --------------------------------------------------------------------------

        # Strobe needs to be high for tWCKTGGL which is 8 tWCK periods
        # NOTE: WCK2CK leveling always happens at 2:1 WCK2CK ratio.
        twcktggl = 4
        wckl_strobe_dly = TappedDelayLine(
            signal = self._wlevel_strobe.re,
            ntaps  = twcktggl
        )
        self.submodules += wckl_strobe_dly

        wckl_strobe_en = Signal()
        self.comb += [
            wckl_strobe_en.eq(reduce(or_, wckl_strobe_dly.taps[0:twcktggl]))
        ]

        wckl_pattern = Signal(8)
        self.comb += [
            wckl_pattern.eq(bitpattern(patterns["disabled"])),
            If(wckl_strobe_en,
                wckl_pattern.eq(bitpattern(patterns["toggle"])),
            )
        ]

        wckl_out = {2: wckl_pattern[::2], 4: wckl_pattern}[wck_ck_ratio]

        wck_pattern_selected = Signal(2*wck_ck_ratio)
        self.comb += [
            If(self._wlevel_en.storage,
                wck_pattern_selected.eq(wckl_out)
            ).Else(
                wck_pattern_selected.eq(wck_out)
            )
        ]

        for byte in range(self.databits//8):
            # output
            self.submodules += BitSlip(
                dw     = 2*wck_ck_ratio,
                cycles = bitslip_cycles,
                rst    = self.get_rst(byte, self._wdly_dq_bitslip_rst.re),
                slp    = self.get_inc(byte, self._wdly_dq_bitslip.re),
                i      = wck_pattern_selected,
                o      = self.out.wck[byte],
            )

        # Write Control Path -----------------------------------------------------------------------
        wrtap = write_latency - 1
        assert wrtap >= 0

        wrdata_en = TappedDelayLine(
            signal = reduce(or_, [dfi.phases[i].wrdata_en for i in range(nphases)]),
            ntaps  = wrtap + (burst_ck_cycles-1) + 2
        )
        self.submodules += wrdata_en

        dq_oe = Signal()
        self.comb += dq_oe.eq(reduce(or_, wrdata_en.taps[wrtap:wrtap+burst_ck_cycles]))

        # Read Control Path ------------------------------------------------------------------------
        rddata_en = TappedDelayLine(
            signal = reduce(or_, [dfi.phases[i].rddata_en for i in range(nphases)]),
            ntaps  = self.settings.read_latency + burst_ck_cycles
        )
        self.submodules += rddata_en

        # Data Path --------------------------------------------------------------------------------
        self.comb += self.out.dq_oe.eq(delayed(self, dq_oe))
        self.comb += self.out.dmi_oe.eq(self.out.dq_oe if masked_write else 0)

        wrdata_ck = Signal(self.settings.dfi_databits//burst_ck_cycles)
        wrdata_hold = HoldValid([("data", self.settings.dfi_databits)])
        wrdata_converter = stream.Converter(
            nbits_from = self.settings.dfi_databits,
            nbits_to   = len(wrdata_ck),
        )
        self.submodules += wrdata_hold, wrdata_converter

        self.comb += [
            wrdata_hold.sink.data.eq(self.dfi.p0.wrdata),
            wrdata_hold.sink.valid.eq(wrdata_en.taps[wrtap]),
            wrdata_hold.source.connect(wrdata_converter.sink),
            wrdata_ck.eq(wrdata_converter.source.data),
            wrdata_converter.source.ready.eq(1),
        ]

        if masked_write:
            wrdata_mask_ck = Signal(len(wrdata_ck)//8)
            wrdata_mask_hold = HoldValid([("data", self.settings.dfi_databits//8)])
            wrdata_mask_converter = stream.Converter(
                nbits_from = self.settings.dfi_databits//8,
                nbits_to   = len(wrdata_mask_ck),
            )
            self.submodules += wrdata_mask_hold, wrdata_mask_converter

            self.comb += [
                wrdata_mask_hold.sink.data.eq(self.dfi.p0.wrdata_mask),
                wrdata_mask_hold.sink.valid.eq(wrdata_en.taps[wrtap]),
                wrdata_mask_hold.source.connect(wrdata_mask_converter.sink),
                wrdata_mask_ck.eq(wrdata_mask_converter.source.data),
                wrdata_mask_converter.source.ready.eq(1),
            ]

        rddata_ck = Signal(self.settings.dfi_databits//burst_ck_cycles)
        rddata_converter = stream.Converter(
            nbits_from = len(rddata_ck),
            nbits_to   = self.settings.dfi_databits,
        )
        self.submodules += rddata_converter

        # TODO: check if that -1 is correct, from Migen simulations it seems like it shouldn't
        # be there, but in LPDDR4 the bitslip_range value was needed
        rddata_start = read_latency - burst_ck_cycles - 1
        self.comb += [
            rddata_converter.sink.data.eq(rddata_ck),
            rddata_converter.sink.valid.eq(reduce(or_, rddata_en.taps[rddata_start:rddata_start+burst_ck_cycles])),
            rddata_converter.source.ready.eq(1),
            self.dfi.p0.rddata.eq(rddata_converter.source.data),
            self.dfi.p0.rddata_valid.eq(rddata_converter.source.valid),
        ]

        # -2 is to take into account the serialization time for wck
        read_wck_latency = cmd_latency + cl + burst_ck_cycles - 2
        write_wck_latency = cmd_latency + cwl + burst_ck_cycles - 2

        self.wck_sync_state = Signal(2)
        self.sync += If(self.adapter.wck_sync != 0,
            wck_sync_done.eq(1),
            self.wck_sync_state.eq(self.adapter.wck_sync)
        ).Elif(self.wck_sync_state == WCKSyncType.RD,
            If(reduce(or_, rddata_en.taps[0:read_wck_latency]) == 0,
                wck_sync_done.eq(0),
                self.wck_sync_state.eq(0b00),
            )
        ).Elif(self.wck_sync_state == WCKSyncType.WR,
            If(reduce(or_, wrdata_en.taps[0:write_wck_latency]) == 0,
                wck_sync_done.eq(0),
                self.wck_sync_state.eq(0b00),
            )
        )

        self.comb += self.adapter.wck_sync_done.eq(wck_sync_done)

        for bit in range(self.databits):
            # output
            wrdata = [wrdata_ck[i * self.databits + bit] for i in range(2*wck_ck_ratio)]
            self.submodules += BitSlip(
                dw     = 2*wck_ck_ratio,
                cycles = bitslip_cycles,
                rst    = self.get_rst(bit//8, wdly_dq_bitslip_rst),
                slp    = self.get_inc(bit//8, wdly_dq_bitslip),
                i      = Cat(*wrdata),
                o      = self.out.dq_o[bit],
            )

            if masked_write and bit % 8 == 0:
                byte = bit//8
                wrdata_mask = [wrdata_mask_ck[i * self.databits//8 + byte] for i in range(2*wck_ck_ratio)]
                self.submodules += BitSlip(
                    dw     = 2*wck_ck_ratio,
                    cycles = bitslip_cycles,
                    rst    = self.get_rst(byte, wdly_dq_bitslip_rst),
                    slp    = self.get_inc(byte, wdly_dq_bitslip),
                    i      = Cat(*wrdata_mask),
                    o      = self.out.dmi_o[byte],
                )

            # input
            dq_i_bs = Signal(2*wck_ck_ratio)
            self.submodules += BitSlip(
                dw     = 2*wck_ck_ratio,
                cycles = bitslip_cycles,
                rst    = self.get_rst(bit//8, rdly_dq_bitslip_rst),
                slp    = self.get_inc(bit//8, rdly_dq_bitslip),
                i      = self.out.dq_i[bit],
                o      = dq_i_bs,
            )
            for i in range(2*wck_ck_ratio):
                self.comb += rddata_ck[i * self.databits + bit].eq(dq_i_bs[i])

    def get_rst(self, byte, rst):
        return (self._dly_sel.storage[byte] & rst) | self._rst.storage

    def get_inc(self, byte, inc):
        return self._dly_sel.storage[byte] & inc



#
# This file is part of LiteDRAM.
#
# Copyright (c) 2016-2021 Florent Kermarrec <florent@enjoy-digital.fr>
# Copyright (c) 2015 Sebastien Bourdeauducq <sb@m-labs.hk>
# Copyright (c) 2018 John Sully <john@csquare.ca>
# Copyright (c) 2016 Tim 'mithro' Ansell <mithro@mithis.com>
# SPDX-License-Identifier: BSD-2-Clause

"""Direct Memory Access (DMA) reader and writer modules."""

from math import log2

from migen import *

from litex.soc.interconnect.csr import *
from litex.soc.interconnect import stream

from litedram.common import LiteDRAMNativePort
from litedram.frontend.axi import LiteDRAMAXIPort

# LiteDRAMDMAReader --------------------------------------------------------------------------------

class LiteDRAMDMAReader(Module, AutoCSR):
    """Read data from DRAM memory.

    For every address written to the sink, one DRAM word will be produced on
    the source.

    Parameters
    ----------
    port : port
        Port on the DRAM memory controller to read from (Native or AXI).

    fifo_depth : int
        How many request results the output FIFO can contain (and thus how many
        read requests can be outstanding at once).

    fifo_buffered : bool
        Implement FIFO in Block Ram.

    Attributes
    ----------
    sink : Record("address")
        Sink for DRAM addresses to be read.

    source : Record("data")
        Source for DRAM word results from reading.

    rsv_level: Signal()
        FIFO reservation level counter
    """

    def __init__(self, port, fifo_depth=16, fifo_buffered=False, with_csr=False):
        assert isinstance(port, (LiteDRAMNativePort, LiteDRAMAXIPort))
        self.port   = port
        self.enable = enable = Signal(reset=1)
        self.sink   = sink   = stream.Endpoint([("address", port.address_width)])
        self.source = source = stream.Endpoint([("data", port.data_width)])

        # # #

        # Native / AXI selection -------------------------------------------------------------------
        is_native = isinstance(port, LiteDRAMNativePort)
        is_axi    = isinstance(port, LiteDRAMAXIPort)
        if is_native:
            (cmd, rdata) = port.cmd, port.rdata
        elif is_axi:
            (cmd, rdata) = port.ar, port.r
        else:
            raise NotImplementedError

        # Request issuance -------------------------------------------------------------------------
        request_enable = Signal()
        request_issued = Signal()

        if is_native:
            self.comb += cmd.we.eq(0)
        if is_axi:
            self.comb += cmd.size.eq(int(log2(port.data_width//8)))
        self.comb += [
            cmd.addr.eq(sink.address),
            cmd.valid.eq(enable & sink.valid & request_enable),
            sink.ready.eq(enable & cmd.ready & request_enable),
            request_issued.eq(cmd.valid & cmd.ready)
        ]

        # FIFO reservation level counter -----------------------------------------------------------
        # - Incremented when data is planned to be queued.
        # - Decremented when data is dequeued.
        data_dequeued = Signal()
        self.rsv_level = rsv_level = Signal(max=fifo_depth+1)
        self.sync += [
            If(request_issued,
                If(~data_dequeued, rsv_level.eq(self.rsv_level + 1))
            ).Elif(data_dequeued,
                rsv_level.eq(rsv_level - 1)
            )
        ]
        self.comb += request_enable.eq(rsv_level != fifo_depth)

        # FIFO -------------------------------------------------------------------------------------
        fifo = stream.SyncFIFO([("data", port.data_width)], fifo_depth, fifo_buffered)
        self.submodules += fifo

        self.comb += [
            rdata.connect(fifo.sink, omit={"id", "resp"}),
            fifo.source.connect(source, omit={"ready"}),
            fifo.source.ready.eq(source.ready | ~enable), # Flush FIFO/Reservation counter when disabled.
            data_dequeued.eq(fifo.source.valid & fifo.source.ready)
        ]

        if with_csr:
            self.add_csr()

    def add_csr(self, default_base=0, default_length=0, default_enable=0, default_loop=0):
        self._base   = CSRStorage(32, reset=default_base)
        self._length = CSRStorage(32, reset=default_length)
        self._enable = CSRStorage(reset=default_enable)
        self._done   = CSRStatus()
        self._loop   = CSRStorage(reset=default_loop)
        self._offset = CSRStatus(32)

        # # #

        shift  = log2_int(self.port.data_width//8)
        base   = Signal(self.port.address_width)
        offset = Signal(self.port.address_width)
        length = Signal(self.port.address_width)
        self.comb += self.enable.eq(self._enable.storage)
        self.comb += base.eq(self._base.storage[shift:])
        self.comb += length.eq(self._length.storage[shift:])

        self.comb += self._offset.status.eq(offset)

        fsm = FSM(reset_state="IDLE")
        fsm = ResetInserter()(fsm)
        self.submodules.fsm = fsm
        self.comb += fsm.reset.eq(~self._enable.storage)
        fsm.act("IDLE",
            NextValue(offset, 0),
            NextState("RUN"),
        )
        fsm.act("RUN",
            self.sink.valid.eq(1),
            self.sink.last.eq(offset == (length - 1)),
            self.sink.address.eq(base + offset),
            If(self.sink.ready,
                NextValue(offset, offset + 1),
                If(self.sink.last,
                    If(self._loop.storage,
                        NextValue(offset, 0)
                    ).Else(
                        NextState("DONE")
                    )
                )
            )
        )
        fsm.act("DONE", self._done.status.eq(1))

# LiteDRAMDMAWriter --------------------------------------------------------------------------------

class LiteDRAMDMAWriter(Module, AutoCSR):
    """Write data to DRAM memory.

    Parameters
    ----------
    port : port
        Port on the DRAM memory controller to write to (Native or AXI).

    fifo_depth : int
        How many requests the input FIFO can contain (and thus how many write
        requests can be outstanding at once).

    fifo_buffered : bool
        Implement FIFO in Block Ram.

    Attributes
    ----------
    sink : Record("address", "data")
        Sink for DRAM addresses and DRAM data word to be written too.
    """
    def __init__(self, port, fifo_depth=16, fifo_buffered=False, with_csr=False):
        assert isinstance(port, (LiteDRAMNativePort, LiteDRAMAXIPort))
        self.port = port
        self.sink = sink = stream.Endpoint([("address", port.address_width),
                                            ("data", port.data_width)])

        # # #

        # Native / AXI selection -------------------------------------------------------------------
        is_native = isinstance(port, LiteDRAMNativePort)
        is_axi    = isinstance(port, LiteDRAMAXIPort)
        if is_native:
            (cmd, wdata) = port.cmd, port.wdata
        elif is_axi:
            (cmd, wdata) = port.aw, port.w
            self.comb += port.b.ready.eq(1) # Always ack write responses.
        else:
            raise NotImplementedError

        # FIFO -------------------------------------------------------------------------------------
        fifo = stream.SyncFIFO([("data", port.data_width)], fifo_depth, fifo_buffered)
        self.submodules += fifo

        if is_native:
            self.comb += cmd.we.eq(1)
        if is_axi:
            self.comb += cmd.size.eq(int(log2(port.data_width//8)))
        self.comb += [
            cmd.addr.eq(sink.address),
            cmd.valid.eq(fifo.sink.ready & sink.valid),
            sink.ready.eq(fifo.sink.ready & cmd.ready),
            fifo.sink.valid.eq(sink.valid & cmd.ready),
            fifo.sink.data.eq(sink.data)
        ]

        if is_native:
            self.comb += wdata.we.eq(2**(port.data_width//8)-1)
        if is_axi:
            self.comb += wdata.strb.eq(2**(port.data_width//8)-1)
        self.comb += [
            wdata.valid.eq(fifo.source.valid),
            fifo.source.ready.eq(wdata.ready),
            wdata.data.eq(fifo.source.data)
        ]

        if with_csr:
            self.add_csr()

    def add_csr(self, default_base=0, default_length=0, default_enable=0, default_loop=0):
        self._sink = self.sink
        self.sink  = stream.Endpoint([("data", self.port.data_width)])

        self._base   = CSRStorage(32, reset=default_base)
        self._length = CSRStorage(32, reset=default_length)
        self._enable = CSRStorage(reset=default_enable)
        self._done   = CSRStatus()
        self._loop   = CSRStorage(reset=default_loop)
        self._offset = CSRStatus(32)

        # # #

        shift  = log2_int(self.port.data_width//8)
        base   = Signal(self.port.address_width)
        offset = Signal(self.port.address_width)
        length = Signal(self.port.address_width)
        self.comb += base.eq(self._base.storage[shift:])
        self.comb += length.eq(self._length.storage[shift:])

        self.comb += self._offset.status.eq(offset)

        fsm = FSM(reset_state="IDLE")
        fsm = ResetInserter()(fsm)
        self.submodules.fsm = fsm
        self.comb += fsm.reset.eq(~self._enable.storage)
        fsm.act("IDLE",
            self.sink.ready.eq(1),
            NextValue(offset, 0),
            NextState("RUN"),
        )
        fsm.act("RUN",
            self._sink.valid.eq(self.sink.valid),
            self._sink.last.eq(offset == (length - 1)),
            self._sink.address.eq(base + offset),
            self._sink.data.eq(self.sink.data),
            self.sink.ready.eq(self._sink.ready),
            If(self.sink.valid & self.sink.ready,
                NextValue(offset, offset + 1),
                If(self._sink.last,
                    If(self._loop.storage,
                        NextValue(offset, 0)
                    ).Else(
                        NextState("DONE")
                    )
                )
            )
        )
        fsm.act("DONE", self._done.status.eq(1))



#!/usr/bin/env python3

# This file is Copyright (c) 2015-2019 Florent Kermarrec <florent@enjoy-digital.fr>
# This file is Copyright (c) 2020 Antmicro <www.antmicro.com>
# License: BSD

# Setup:
# https://user-images.githubusercontent.com/1450143/165936416-faebe814-e727-44e5-8392-72931f2abb68.JPG
# - Replace Arty's DDR3 with EM6GA16L.
# - Connect Arty's I2C to PMIC's I2C.

# Build/Use:
# ./arty.py --build --load
# litex_server --udp
# litex_term crossover

import os
import argparse

from migen import *

from litex_boards.platforms import digilent_arty
from litex.build.generic_platform import *
from litex.build.xilinx.vivado import vivado_build_args, vivado_build_argdict

from litex.soc.cores.clock import *
from litex.soc.integration.soc_core import *
from litex.soc.integration.builder import *
from litex.soc.cores.led import LedChaser
from litex.soc.cores.bitbang import I2CMaster
from litex.soc.interconnect.csr import AutoCSR, CSRStatus, CSRStorage

from litedram.init import get_sdram_phy_py_header
from litedram.core.controller import ControllerSettings
from litedram.phy.rpc.basephy import RPCPads
from litedram.phy.rpc.s7phy import A7RPCPHY
from litedram.modules import EM6GA16L

from liteeth.phy.mii import LiteEthPHYMII

from litescope import LiteScopeAnalyzer

# Pads ---------------------------------------------------------------------------------------------

class RPCPadsDDR3(RPCPads):
    def map(self, pads):
        self.clk_p  = pads.clk_p
        self.clk_n  = pads.clk_n
        self.cs_n   = pads.dm[1]
        self.dqs_p  = Array([pads.dqs_p[1], pads.dqs_p[0]])
        self.dqs_n  = Array([pads.dqs_n[1], pads.dqs_n[0]])
        self.stb    = pads.dm[0]
        self.db     = Array([
            pads.dq[10],
            pads.dq[14],
            pads.dq[8],
            pads.dq[12],
            pads.dq[15],
            pads.dq[9],
            pads.dq[13],
            pads.dq[11],
            pads.dq[2],
            pads.dq[6],
            pads.dq[0],
            pads.dq[4],
            pads.dq[7],
            pads.dq[1],
            pads.dq[5],
            pads.dq[3],
        ])


def ddram_dbg_pmod_io(pmod_db, pmod_others, len_db=1):
    unused = "V15 U16 P14 T11 R12 T14 T15 T16 N15 M16 V17 U18 R17 P17 U11 V16 M13 R10 R11 R13 R15 P15 R16".split(" ")
    pmod_db_pins = [f"{pmod_db}:{i}" for i in range(8)]
    return [
        ("ddram_dbg", 0,
            Subsignal("clk_p", Pins(f"{pmod_others}:0")),
            Subsignal("clk_n", Pins(f"{pmod_others}:4")),
            Subsignal("dqs_p", Pins(f"{pmod_others}:1 {pmod_others}:5")),
            Subsignal("dqs_n", Pins(f"{pmod_others}:6 {pmod_others}:7")),
            Subsignal("cs_n",  Pins(f"{pmod_others}:2")),
            Subsignal("stb",   Pins(f"{pmod_others}:3")),
            Subsignal("db",    Pins(*(pmod_db_pins[:len_db] + unused[:16 - len_db]))),
            IOStandard("LVCMOS33")
        ),
        ("dbg_pmod", 0, Pins(*pmod_db_pins[len_db:]), IOStandard("LVCMOS33")),
    ]

_ddram_dbg_io = ddram_dbg_pmod_io("pmoda", "pmodd")

def ddram_io():
    return [
        ("ddram_sstl15", 0,
            Subsignal("a", Pins(
                "R2 M6 N4 T1 N6 R7 V6 U7",
                "R8 V7 R6 U6 T6 T8"),
                IOStandard("SSTL15")),
            Subsignal("ba",    Pins("R1 P4 P2"), IOStandard("SSTL15")),
            Subsignal("ras_n", Pins("P3"), IOStandard("SSTL15")),
            Subsignal("cas_n", Pins("M4"), IOStandard("SSTL15")),
            Subsignal("we_n",  Pins("P5"), IOStandard("SSTL15")),
            Subsignal("cs_n",  Pins("U8"), IOStandard("SSTL15")),
            Subsignal("dm", Pins("L1 U1"), IOStandard("SSTL15")),
            Subsignal("dq", Pins(
                "K5 L3 K3 L6 M3 M1 L4 M2",
                "V4 T5 U4 V5 V1 T3 U3 R3"),
                IOStandard("SSTL15"),
                Misc("IN_TERM=UNTUNED_SPLIT_40")),
            Subsignal("dqs_p", Pins("N2 U2"),
                IOStandard("DIFF_SSTL15"),
                Misc("IN_TERM=UNTUNED_SPLIT_40")),
            Subsignal("dqs_n", Pins("N1 V2"),
                IOStandard("DIFF_SSTL15"),
                Misc("IN_TERM=UNTUNED_SPLIT_40")),
            Subsignal("clk_p", Pins("U9"), IOStandard("DIFF_SSTL15")),
            Subsignal("clk_n", Pins("V9"), IOStandard("DIFF_SSTL15")),
            Subsignal("cke",   Pins("N5"), IOStandard("SSTL15")),
            Subsignal("odt",   Pins("R5"), IOStandard("SSTL15")),
            Subsignal("reset_n", Pins("K6"), IOStandard("SSTL15")),
            Misc("SLEW=FAST"),
        )
    ]

# CRG ----------------------------------------------------------------------------------------------

class _CRG(Module, AutoCSR):
    def __init__(self, platform, sys_clk_freq, dynamic=False):
        self.clock_domains.cd_sys       = ClockDomain()
        self.clock_domains.cd_clk200    = ClockDomain()
        self.clock_domains.cd_sys4x_90  = ClockDomain(reset_less=True)
        self.clock_domains.cd_sys4x_180 = ClockDomain(reset_less=True)
        if dynamic:
            self.clock_domains.cd_sys_pll  = ClockDomain()
            self.clock_domains.cd_uart      = ClockDomain()
        else:
            self.clock_domains.cd_eth      = ClockDomain()

        # # #

        def add_sys_clocks(pll):
            pll.create_clkout(self.cd_sys,       sys_clk_freq)
            pll.create_clkout(self.cd_sys4x_90,  4*sys_clk_freq, phase=90)
            pll.create_clkout(self.cd_sys4x_180, 4*sys_clk_freq, phase=180)

        self.submodules.main_pll = main_pll = S7PLL(speedgrade=-1)
        self.comb += main_pll.reset.eq(~platform.request("cpu_reset"))
        main_pll.register_clkin(platform.request("clk100"), 100e6)
        if dynamic:
            main_pll.create_clkout(self.cd_sys_pll, sys_clk_freq)
            main_pll.create_clkout(self.cd_clk200, 200e6)
            main_pll.create_clkout(self.cd_uart,   100e6)
            main_pll.expose_drp()
        else:
            add_sys_clocks(main_pll)
            main_pll.create_clkout(self.cd_clk200, 200e6)
            main_pll.create_clkout(self.cd_eth,     25e6)

        if dynamic:
            self.submodules.sys_pll = sys_pll = S7PLL(speedgrade=-1)
            self.comb += sys_pll.reset.eq(~main_pll.locked)
            sys_pll.register_clkin(self.cd_sys_pll.clk, sys_clk_freq)
            add_sys_clocks(sys_pll)

            # a simple way to approximately check the actual frequency
            ref_clk_counter      = Signal(32)
            sys_clk_counter      = Signal(32)
            self.counters_run    = CSRStorage()
            self.ref_clk_counter = CSRStatus(32)
            self.sys_clk_counter = CSRStatus(32)
            self.sync.clk200 += If(self.counters_run.storage, ref_clk_counter.eq(ref_clk_counter + 1))
            self.sync.sys    += If(self.counters_run.storage, sys_clk_counter.eq(sys_clk_counter + 1))
            self.comb += [
                self.ref_clk_counter.status.eq(ref_clk_counter),
                self.sys_clk_counter.status.eq(sys_clk_counter),
            ]

        self.submodules.idelayctrl = S7IDELAYCTRL(self.cd_clk200)
        if not dynamic:
            self.comb += platform.request("eth_ref_clk").eq(self.cd_eth.clk)

# BaseSoC ------------------------------------------------------------------------------------------

class BaseSoC(SoCCore):
    def __init__(self, sys_clk_freq=int(100e6), ip_address="192.168.1.50", debug_pmod=False,
                 no_sdram_init=False, dynamic_freq=False, with_analyzer=False, **kwargs):
        platform = digilent_arty.Platform()

        # SoCCore ----------------------------------------------------------------------------------
        kwargs["integrated_rom_size"] = 0xa000
        kwargs["uart_name"]           = "crossover"
        SoCCore.__init__(self, platform, sys_clk_freq,
            ident               = "LiteX SoC on Arty A7",
            integrated_rom_mode = "rw", # to allow reloading BIOS
            **kwargs)

        # CRG --------------------------------------------------------------------------------------
        self.submodules.crg = _CRG(platform, sys_clk_freq, dynamic=dynamic_freq)

        # DDR3 SDRAM -------------------------------------------------------------------------------
        platform.add_extension(ddram_io())
        platform.add_platform_command("set_property INTERNAL_VREF 0.750 [get_iobanks 34]")
        if debug_pmod:
            from pprint import pprint
            print()
            print("  PMOD placement:")
            print("  ---------------")
            pprint(_ddram_dbg_io)
            print()
            platform.add_extension(_ddram_dbg_io)
            ddram_pads = RPCPads(platform.request("ddram_dbg"))
            dbg_pmod = platform.request("dbg_pmod")
            self.comb += [
                dbg_pmod[0].eq(ClockSignal()),
                dbg_pmod[1].eq(ClockSignal("sys4x_90")),
                dbg_pmod[2].eq(ClockSignal("sys4x_180")),
            ]
        else:
            ddram_pads = RPCPadsDDR3(platform.request("ddram_sstl15"))

        self.submodules.ddrphy = A7RPCPHY(pads=ddram_pads,
                                          sys_clk_freq=100e6 if debug_pmod else sys_clk_freq,
                                          iodelay_clk_freq=200e6)
                                          # no_differential=debug_pmod)
        module = EM6GA16L(100e6 if debug_pmod else sys_clk_freq, "1:4")

        self.ddrphy.refresh_enable = CSRStorage()

        controller_settings = ControllerSettings()
        controller_settings.auto_precharge = False
        controller_settings.with_refresh = self.ddrphy.refresh_enable.storage

        self.add_sdram("sdram",
            phy                 = self.ddrphy,
            module              = module,
            controller_settings = controller_settings,
        )

        self.add_constant("SET_DDR_VCC_15")
        # self.add_constant("RPC_UTR_TEST")

        # make BIOS wait for manual initialization from RemoteClient
        # (waits for the given value in ctrl_scratch)
        # self.add_constant("SDRAM_WAIT_MANUAL_INIT_SEQUENCE", 0xc001c0de)

        if no_sdram_init:
            self.add_constant("SDRAM_INIT_DISABLE")

        # I2C --------------------------------------------------------------------------------------
        # Used to configure the power management chip (DA9062) on the board to use different DDRVCC.
        # We explicitly use tristate SCL, because the chip operates on 5V logic (pull ups are
        # through 100k resistors so shouldn't damage FPGA pins).
        self.submodules.i2c = I2CMaster(platform.request("i2c"))
        self.i2c.add_init(addr=0x58, init=[
            # Vbuck2A / 1.5V.
            (0xa3, 0x78),
            # Vbuck2B / 1.5V.
            (0xb4, 0x78),
        ])

        if dynamic_freq:
            # UartBone -----------------------------------------------------------------------------
            self.add_uartbone(name="serial", clk_freq=100e6, baudrate=1e6, cd="uart")
        else:
            # Etherbone ----------------------------------------------------------------------------
            self.submodules.ethphy = LiteEthPHYMII(
                clock_pads = self.platform.request("eth_clocks"),
                pads       = self.platform.request("eth"))
            self.add_etherbone(phy=self.ethphy, ip_address=ip_address)

        # Leds -------------------------------------------------------------------------------------
        self.submodules.leds = LedChaser(
            pads         = Cat(*[platform.request("user_led", i) for i in range(4)]),
            sys_clk_freq = sys_clk_freq)

        # Analyzer ---------------------------------------------------------------------------------

        if with_analyzer:
            self.ddrphy.finalize()
            analyzer_signals = [
                *[self.ddrphy.dfi.phases[p].cas_n for p in range(self.ddrphy.nphases)],
                *[self.ddrphy.dfi.phases[p].ras_n for p in range(self.ddrphy.nphases)],
                *[self.ddrphy.dfi.phases[p].we_n  for p in range(self.ddrphy.nphases)],
                *[self.ddrphy.dfi.phases[p].reset_n  for p in range(self.ddrphy.nphases)],
                self.ddrphy.stb_1ck_in,
                # *self.ddrphy.db_1ck_out,
                # *self.ddrphy.db_1ck_in,
                self.ddrphy.stb_1ck_out,
                self.ddrphy.dqs_1ck_out,
                self.ddrphy.dqs_1ck_in,
                self.ddrphy.cs_n_1ck_out,
                self.ddrphy.clk_1ck_out,
                self.ddrphy.dq_data_en,
                self.ddrphy.dq_mask_en,
                self.ddrphy.dq_cmd_en,
                self.ddrphy.dq_read_stb,
                self.ddrphy.dq_in_cnt,
                self.ddrphy.db_cnt,
                self.ddrphy.dqs_cnt,
                self.ddrphy.rddata_en.sr,
                self.ddrphy.wrdata_en.sr,
                self.ddrphy.db_oe,
                self.ddrphy.dqs_oe,
                self.ddrphy.reset_fsm.state,
            ]
            self.submodules.analyzer = LiteScopeAnalyzer(analyzer_signals,
                depth        = 512,
                register     = True,
                clock_domain = "sys",
                csr_csv      = "analyzer.csv")

        self.add_constant("SDRAM_DEBUG")

        # Print info
        def dump(obj):
            print()
            print(" " + obj.__class__.__name__)
            print(" " + "-" * len(obj.__class__.__name__))
            d = obj if isinstance(obj, dict) else vars(obj)
            for var, val in d.items():
                if var == "self":
                    continue
                print("  {}: {}".format(var, val))
        print("=" * 80)
        dump(self.ddrphy.settings)
        dump(module.geom_settings)
        dump(module.timing_settings)
        print()
        print("=" * 80)
        print()
        print("  VCO freq = ", self.crg.main_pll.compute_config()["vco"])
        print()
        print("=" * 80)

    def generate_sdram_phy_py_header(self):
        f = open("sdram_init.py", "w")
        f.write(get_sdram_phy_py_header(
            self.sdram.controller.settings.phy,
            self.sdram.controller.settings.timing))
        f.close()

# Build --------------------------------------------------------------------------------------------

def reboot(wb):
    wb.regs.ctrl_reset.write(1)

def load_rom(wb, filename):
    from litex.soc.integration.common import get_mem_data
    rom_data = get_mem_data(filename, endianness="little")
    print(f"load bios from: {filename} starting at 0x{wb.mems.rom.base:08x}")
    for i, data in enumerate(rom_data):
        wb.write(wb.mems.rom.base + 4*i, data)

def main():
    parser = argparse.ArgumentParser(description="LiteX SoC on Arty A7")
    parser.add_argument("--build",           action="store_true",    help="Build bitstream")
    parser.add_argument("--load",            action="store_true",    help="Load bitstream")
    parser.add_argument("--load-bios",       action="store_true",    help="Load BIOS over Etherbone and reboot SoC")
    builder_args(parser)
    soc_core_args(parser)
    vivado_build_args(parser)
    parser.add_argument("--ip-address",      default="192.168.1.50", help="Use given IP address")
    parser.add_argument("--with-spi-sdcard", action="store_true",    help="Enable SPI-mode SDCard support")
    parser.add_argument("--with-sdcard",     action="store_true",    help="Enable SDCard support")
    parser.add_argument("--debug-pmod",      action="store_true",    help="Send DRAM signals on PMODs (A and D)")
    parser.add_argument("--sys-clk-freq",    default="100e6",        help="Set frequency of the system clock (default 100 MHz)")
    parser.add_argument("--dynamic-freq",    action="store_true",    help="Dynamic system frequency")
    parser.add_argument("--no-sdram-init",   action="store_true",    help="Disable automatic DRAM initialization in BIOS")
    parser.add_argument("--no-analyzer",     action="store_true",    help="Do not add LiteScope analyzer")
    args = parser.parse_args()

    soc = BaseSoC(ip_address=args.ip_address, dynamic_freq=args.dynamic_freq,
                  debug_pmod=args.debug_pmod, sys_clk_freq=int(float(args.sys_clk_freq)),
                  no_sdram_init=args.no_sdram_init, with_analyzer=not args.no_analyzer,
                  **soc_core_argdict(args))
    assert not (args.with_spi_sdcard and args.with_sdcard)
    soc.platform.add_extension(digilent_arty._sdcard_pmod_io)
    if args.with_spi_sdcard:
        soc.add_spi_sdcard()
    if args.with_sdcard:
        soc.add_sdcard()
    builder_kwargs = builder_argdict(args)
    builder_kwargs["csr_csv"] = "csr.csv"
    builder = Builder(soc, **builder_kwargs)
    builder_kwargs = vivado_build_argdict(args)
    builder.build(**builder_kwargs, run=args.build)

    soc.generate_sdram_phy_py_header()

    if args.load:
        prog = soc.platform.create_programmer()
        prog.load_bitstream(os.path.join(builder.gateware_dir, soc.build_name + ".bit"))

    if args.load_bios:
        from litex import RemoteClient
        from litex.soc.integration.common import get_mem_data

        def memwrite(wb, data, base=0x40000000, burst=0xff):
            for i in range(0, len(data), burst):
                wb.write(base + 4 * i, data[i:i + burst])

        wb = RemoteClient()
        wb.open()

        bios_bin = os.path.join(builder.software_dir, "bios", "bios.bin")
        rom_data = get_mem_data(bios_bin, "little")
        print(f"Loading BIOS from: {bios_bin} starting at 0x{wb.mems.rom.base:08x} ...")

        print('Stopping CPU')
        wb.regs.ctrl_reset.write(0b10)  # cpu_rst

        memwrite(wb, rom_data, base=wb.mems.rom.base, burst=1)
        wb.read(wb.mems.rom.base)

        print('Rebooting CPU')
        wb.regs.ctrl_reset.write(0)

        wb.close()

if __name__ == "__main__":
    main()



#
# This file is part of LiteDRAM.
#
# Copyright (c) 2015 Sebastien Bourdeauducq <sb@m-labs.hk>
# Copyright (c) 2016-2019 Florent Kermarrec <florent@enjoy-digital.fr>
# SPDX-License-Identifier: BSD-2-Clause

"""LiteDRAM BankMachine (Rows/Columns management)."""

import math

from migen import *

from litex.soc.interconnect import stream

from litedram.common import *
from litedram.core.multiplexer import *

# AddressSlicer ------------------------------------------------------------------------------------

class _AddressSlicer:
    """Helper for extracting row/col from address

    Column occupies lower bits of the address, row - higher bits. Address has
    a forced alignment, so column does not contain alignment bits.
    """
    def __init__(self, colbits, address_align):
        self.colbits       = colbits
        self.address_align = address_align

    def row(self, address):
        split = self.colbits - self.address_align
        return address[split:]

    def col(self, address):
        split = self.colbits - self.address_align
        return Cat(Replicate(0, self.address_align), address[:split])

# BankMachine --------------------------------------------------------------------------------------

class BankMachine(Module):
    """Converts requests from ports into DRAM commands

    BankMachine abstracts single DRAM bank by keeping track of the currently
    selected row. It converts requests from LiteDRAMCrossbar to targetted
    to that bank into DRAM commands that go to the Multiplexer, inserting any
    needed activate/precharge commands (with optional auto-precharge). It also
    keeps track and enforces some DRAM timings (other timings are enforced in
    the Multiplexer).

    BankMachines work independently from the data path (which connects
    LiteDRAMCrossbar with the Multiplexer directly).

    Stream of requests from LiteDRAMCrossbar is being queued, so that reqeust
    can be "looked ahead", and auto-precharge can be performed (if enabled in
    settings).

    Lock (cmd_layout.lock) is used to synchronise with LiteDRAMCrossbar. It is
    being held when:
     - there is a valid command awaiting in `cmd_buffer_lookahead` - this buffer
       becomes ready simply when the next data gets fetched to the `cmd_buffer`
     - there is a valid command in `cmd_buffer` - `cmd_buffer` becomes ready
       when the BankMachine sends wdata_ready/rdata_valid back to the crossbar

    Parameters
    ----------
    n : int
        Bank number
    address_width : int
        LiteDRAMInterface address width
    address_align : int
        Address alignment depending on burst length
    nranks : int
        Number of separate DRAM chips (width of chip select)
    settings : ControllerSettings
        LiteDRAMController settings

    Attributes
    ----------
    req : Record(cmd_layout)
        Stream of requests from LiteDRAMCrossbar
    refresh_req : Signal(), in
        Indicates that refresh needs to be done, connects to Refresher.cmd.valid
    refresh_gnt : Signal(), out
        Indicates that refresh permission has been granted, satisfying timings
    cmd : Endpoint(cmd_request_rw_layout)
        Stream of commands to the Multiplexer
    """
    def __init__(self, n, address_width, address_align, nranks, settings):
        self.req = req = Record(cmd_layout(address_width))
        self.refresh_req = refresh_req = Signal()
        self.refresh_gnt = refresh_gnt = Signal()

        a  = settings.geom.addressbits
        ba = settings.geom.bankbits + log2_int(nranks)
        self.cmd = cmd = stream.Endpoint(cmd_request_rw_layout(a, ba))

        # # #

        auto_precharge = Signal()

        # Command buffer ---------------------------------------------------------------------------
        cmd_buffer_layout    = [("we", 1), ("addr", len(req.addr))]
        cmd_buffer_lookahead = stream.SyncFIFO(
            cmd_buffer_layout, settings.cmd_buffer_depth,
            buffered=settings.cmd_buffer_buffered)
        cmd_buffer = stream.Buffer(cmd_buffer_layout) # 1 depth buffer to detect row change
        self.submodules += cmd_buffer_lookahead, cmd_buffer
        self.comb += [
            req.connect(cmd_buffer_lookahead.sink, keep={"valid", "ready", "we", "addr"}),
            cmd_buffer_lookahead.source.connect(cmd_buffer.sink),
            cmd_buffer.source.ready.eq(req.wdata_ready | req.rdata_valid),
            req.lock.eq(cmd_buffer_lookahead.source.valid | cmd_buffer.source.valid),
        ]

        slicer = _AddressSlicer(settings.geom.colbits, address_align)

        # Row tracking -----------------------------------------------------------------------------
        row        = Signal(settings.geom.rowbits)
        row_opened = Signal()
        row_hit    = Signal()
        row_open   = Signal()
        row_close  = Signal()
        self.comb += row_hit.eq(row == slicer.row(cmd_buffer.source.addr))
        self.sync += \
            If(row_close,
                row_opened.eq(0)
            ).Elif(row_open,
                row_opened.eq(1),
                row.eq(slicer.row(cmd_buffer.source.addr))
            )

        # Address generation -----------------------------------------------------------------------
        row_col_n_addr_sel = Signal()
        self.comb += [
            cmd.ba.eq(n),
            If(row_col_n_addr_sel,
                cmd.a.eq(slicer.row(cmd_buffer.source.addr))
            ).Else(
                cmd.a.eq((auto_precharge << 10) | slicer.col(cmd_buffer.source.addr))
            )
        ]

        # tWTP (write-to-precharge) controller -----------------------------------------------------
        write_latency = math.ceil(settings.phy.cwl / settings.phy.nphases)
        precharge_time = write_latency + settings.timing.tWR + settings.timing.tCCD # AL=0
        self.submodules.twtpcon = twtpcon = tXXDController(precharge_time)
        self.comb += twtpcon.valid.eq(cmd.valid & cmd.ready & cmd.is_write)

        # tRC (activate-activate) controller -------------------------------------------------------
        self.submodules.trccon = trccon = tXXDController(settings.timing.tRC)
        self.comb += trccon.valid.eq(cmd.valid & cmd.ready & row_open)

        # tRAS (activate-precharge) controller -----------------------------------------------------
        self.submodules.trascon = trascon = tXXDController(settings.timing.tRAS)
        self.comb += trascon.valid.eq(cmd.valid & cmd.ready & row_open)

        # Auto Precharge generation ----------------------------------------------------------------
        # generate auto precharge when current and next cmds are to different rows
        if settings.with_auto_precharge:
            self.comb += \
                If(cmd_buffer_lookahead.source.valid & cmd_buffer.source.valid,
                    If(slicer.row(cmd_buffer_lookahead.source.addr) !=
                       slicer.row(cmd_buffer.source.addr),
                        auto_precharge.eq(row_close == 0)
                    )
                )

        # Control and command generation FSM -------------------------------------------------------
        # Note: tRRD, tFAW, tCCD, tWTR timings are enforced by the multiplexer
        self.submodules.fsm = fsm = FSM()
        fsm.act("REGULAR",
            If(refresh_req,
                NextState("REFRESH")
            ).Elif(cmd_buffer.source.valid,
                If(row_opened,
                    If(row_hit,
                        cmd.valid.eq(1),
                        If(cmd_buffer.source.we,
                            req.wdata_ready.eq(cmd.ready),
                            cmd.is_write.eq(1),
                            cmd.we.eq(1),
                        ).Else(
                            req.rdata_valid.eq(cmd.ready),
                            cmd.is_read.eq(1)
                        ),
                        cmd.cas.eq(1),
                        If(cmd.ready & auto_precharge,
                           NextState("AUTOPRECHARGE")
                        )
                    ).Else(  # row_opened & ~row_hit
                        NextState("PRECHARGE")
                    )
                ).Else(  # ~row_opened
                    NextState("ACTIVATE")
                )
            )
        )
        fsm.act("PRECHARGE",
            # Note: we are presenting the column address, A10 is always low
            If(twtpcon.ready & trascon.ready,
                cmd.valid.eq(1),
                If(cmd.ready,
                    NextState("TRP")
                ),
                cmd.ras.eq(1),
                cmd.we.eq(1),
                cmd.is_cmd.eq(1)
            ),
            row_close.eq(1)
        )
        fsm.act("AUTOPRECHARGE",
            If(twtpcon.ready & trascon.ready,
                NextState("TRP")
            ),
            row_close.eq(1)
        )
        fsm.act("ACTIVATE",
            If(trccon.ready,
                row_col_n_addr_sel.eq(1),
                row_open.eq(1),
                cmd.valid.eq(1),
                cmd.is_cmd.eq(1),
                If(cmd.ready,
                    NextState("TRCD")
                ),
                cmd.ras.eq(1)
            )
        )
        fsm.act("REFRESH",
            If(twtpcon.ready,
                refresh_gnt.eq(1),
            ),
            row_close.eq(1),
            cmd.is_cmd.eq(1),
            If(~refresh_req,
                NextState("REGULAR")
            )
        )
        fsm.delayed_enter("TRP", "ACTIVATE", settings.timing.tRP - 1)
        fsm.delayed_enter("TRCD", "REGULAR", settings.timing.tRCD - 1)



#
# This file is part of LiteDRAM.
#
# Copyright (c) 2021 Antmicro <www.antmicro.com>
# SPDX-License-Identifier: BSD-2-Clause

from operator import or_
from functools import reduce

from migen import *

from litex.soc.interconnect.csr import *

from litedram.common import *
from litedram.phy.dfi import *

from litedram.phy.utils import (bitpattern, delayed, DQSPattern, Serializer, Deserializer, Latency,
    CommandsPipeline)
from litedram.phy.lpddr4.commands import DFIPhaseAdapter


class LPDDR4Output:
    """Unserialized output of LPDDR4PHY. Has to be serialized by concrete implementation."""
    def __init__(self, nphases, databits):
        # Pads: RESET_N, CS, CKE, CK, CA[5:0], DMI[1:0], DQ[15:0], DQS[1:0], ODT_CA
        self.clk     = Signal(2*nphases)
        self.cke     = Signal(nphases)
        self.odt     = Signal(nphases)
        self.reset_n = Signal(nphases)
        self.cs      = Signal(nphases)
        self.ca      = [Signal(nphases)   for _ in range(6)]
        self.dmi_o   = [Signal(2*nphases) for _ in range(databits//8)]
        self.dmi_i   = [Signal(2*nphases) for _ in range(databits//8)]
        self.dmi_oe  = Signal()  # no serialization
        self.dq_o    = [Signal(2*nphases) for _ in range(databits)]
        self.dq_i    = [Signal(2*nphases) for _ in range(databits)]
        self.dq_oe   = Signal()  # no serialization
        self.dqs_o   = [Signal(2*nphases) for _ in range(databits//8)]
        self.dqs_i   = [Signal(2*nphases) for _ in range(databits//8)]
        self.dqs_oe  = Signal()  # no serialization


class LPDDR4PHY(Module, AutoCSR):
    """Core of LPDDR4 PHYs.

    This class implements all the logic required to convert DFI to/from pads.
    It works in a single clock domain. Signals for DRAM pads are stored in
    LPDDR4Output (self.out). Concrete implementations of LPDDR4 PHYs derive
    from this class and perform (de-)serialization of LPDDR4Output to pads.

    DFI commands
    ------------
    Not all LPDDR4 commands map directly to DFI commands. For this reason ZQC
    is treated specially in that DFI ZQC is translated into LPDDR4 MPC and has
    different interpretation depending on DFI.address.

    Due to the fact that LPDDR4 has 256-bit Mode Register space, the DFI MRS
    command encodes both register address *and* value in DFI.address (instead
    of the default in LiteDRAM to split them between DFI.address and DFI.bank).
    The MRS command is used both for Mode Register Write and Mode Register Read.
    The command is selected based on the value of DFI.bank.

    Refer to the documentation in `commands.py` for further information.

    Parameters
    ----------
    pads : object
        Object containing LPDDR4 pads.
    sys_clk_freq : float
        Frequency of memory controller's clock.
    ser_latency : Latency
        Additional latency introduced due to signal serialization.
    des_latency : Latency
        Additional latency introduced during signal deserialization.
    phytype : str
        Name of the PHY (concrete implementation).
    cmd_delay : int
        Used to force cmd delay during initialization in BIOS.
    masked_write : bool
        Use MASKED-WRITE commands if True else use WRITE commands (data masking will not work).
    extended_overlaps_check : bool
        Include additional command overlap checks. Makes no sense during normal operation
        (when the DRAM controller works correctly), so use `False` to avoid wasting resources.
    """
    def __init__(self, pads, *,
                 sys_clk_freq, ser_latency, des_latency, phytype,
                 cmd_delay=None, masked_write=True, extended_overlaps_check=False):
        self.pads        = pads
        self.memtype     = memtype     = "LPDDR4"
        self.nranks      = nranks      = 1 if not hasattr(pads, "cs_n") else len(pads.cs_n)
        self.databits    = databits    = len(pads.dq)
        self.addressbits = addressbits = 17  # for activate row address
        self.bankbits    = bankbits    = 6  # 3 bankbits, but we use 6 for Mode Register address in MRS
        self.nphases     = nphases     = 8
        self.tck         = tck         = 1 / (nphases*sys_clk_freq)
        assert databits % 8 == 0

        # Parameters -------------------------------------------------------------------------------
        def get_cl_cw(memtype, tck):
            # MT53E256M16D1, No DBI, Set A
            f_to_cl_cwl = OrderedDict()
            f_to_cl_cwl[ 532e6] = ( 6,  4)
            f_to_cl_cwl[1066e6] = (10,  6)
            f_to_cl_cwl[1600e6] = (14,  8)
            f_to_cl_cwl[2132e6] = (20, 10)
            f_to_cl_cwl[2666e6] = (24, 12)
            f_to_cl_cwl[3200e6] = (28, 14)
            f_to_cl_cwl[3732e6] = (32, 16)
            f_to_cl_cwl[4266e6] = (36, 18)
            for f, (cl, cwl) in f_to_cl_cwl.items():
                if tck >= 2/f:
                    return cl, cwl
            raise ValueError

        # Bitslip introduces latency from 1 up to `cycles + 1`
        # FIXME: (check if True) from tests on hardware it seems we need 1 more cycle
        #   of read_latency, probably to have space for manipulating bitslip values
        bitslip_cycles  = 1
        bitslip_range   = 1
        # Commands are sent over 4 DRAM clocks (sys8x) and we count cl/cwl from last bit
        cmd_latency     = 4
        # Commands read from adapters are delayed on ConstBitSlips
        ca_latency      = 1

        cl, cwl         = get_cl_cw(memtype, tck)
        cl_sys_latency  = get_sys_latency(nphases, cl)
        cwl_sys_latency = get_sys_latency(nphases, cwl)
        # For reads we need to account for ser+des latency to make sure we get the data in-phase with sys clock
        rdphase = get_sys_phase(nphases, cl_sys_latency, cl + cmd_latency + ser_latency.sys8x % 8 + des_latency.sys8x % 8)
        # No need to modify wrphase, because ser_latency applies the same to both CA and DQ
        wrphase = get_sys_phase(nphases, cwl_sys_latency, cwl + cmd_latency)

        # When the calculated phase is negative, it means that we need to increase sys latency
        def updated_latency(phase, sys_latency):
            while phase < 0:
                phase += nphases
                sys_latency += 1
            return phase, sys_latency

        wrphase, cwl_sys_latency = updated_latency(wrphase, cwl_sys_latency)
        rdphase, cl_sys_latency = updated_latency(rdphase, cl_sys_latency)

        # Read latency
        read_data_delay = ca_latency + ser_latency.sys8x//8 + cl_sys_latency  # DFI cmd -> read data on DQ
        read_des_delay  = des_latency.sys8x//8 + bitslip_cycles+bitslip_range  # data on DQ -> data on DFI rddata
        read_latency    = read_data_delay + read_des_delay

        # Write latency
        write_latency = cwl_sys_latency

        # Registers --------------------------------------------------------------------------------
        self._rst             = CSRStorage()

        self._wlevel_en     = CSRStorage()
        self._wlevel_strobe = CSR()

        self._dly_sel = CSRStorage(databits//8)

        self._rdly_dq_bitslip_rst = CSR()
        self._rdly_dq_bitslip     = CSR()

        self._wdly_dq_bitslip_rst = CSR()
        self._wdly_dq_bitslip     = CSR()

        self._rdphase = CSRStorage(log2_int(nphases), reset=rdphase)
        self._wrphase = CSRStorage(log2_int(nphases), reset=wrphase)

        # PHY settings -----------------------------------------------------------------------------
        self.settings = PhySettings(
            phytype       = phytype,
            memtype       = memtype,
            databits      = databits,
            dfi_databits  = 2*databits,
            nranks        = nranks,
            nphases       = nphases,
            rdphase       = self._rdphase.storage,
            wrphase       = self._wrphase.storage,
            cl            = cl,
            cwl           = cwl,
            read_latency  = read_latency,
            write_latency = write_latency,
            cmd_latency   = cmd_latency,
            cmd_delay     = cmd_delay,
            bitslips      = 16,
        )

        # DFI Interface ----------------------------------------------------------------------------
        # Due to the fact that LPDDR4 has 16n prefetch we use 8 phases to be able to read/write a
        # whole burst during a single controller clock cycle. PHY should use sys8x clock.
        self.dfi = dfi = Interface(addressbits, bankbits, nranks, 2*databits, nphases=8)

        # # #

        adapters = [DFIPhaseAdapter(phase, masked_write=masked_write) for phase in self.dfi.phases]
        self.submodules += adapters

        # Now prepare the data by converting the sequences on adapters into sequences on the pads.
        # We have to ignore overlapping commands, and module timings have to ensure that there are
        # no overlapping commands anyway.
        self.out = LPDDR4Output(nphases, databits)

        # Clocks -----------------------------------------------------------------------------------
        self.comb += self.out.clk.eq(bitpattern("-_-_-_-_" * 2))

        # Simple commands --------------------------------------------------------------------------
        self.comb += [
            self.out.cke.eq(Cat(delayed(self, phase.cke) for phase in self.dfi.phases)),
            self.out.odt.eq(Cat(delayed(self, phase.odt) for phase in self.dfi.phases)),
            self.out.reset_n.eq(Cat(delayed(self, phase.reset_n) for phase in self.dfi.phases)),
        ]

        # LPDDR4 Commands --------------------------------------------------------------------------
        # Each LPDDR4 command can span several phases (2 or 4), so in theory the commands could
        # overlap. No overlap should be guaranteed by the controller based on module timings, but
        # we also include an overlaps check in PHY logic.
        self.submodules.commands = CommandsPipeline(adapters,
            cs_ser_width = len(self.out.cs),
            ca_ser_width = len(self.out.ca[0]),
            ca_nbits     = len(self.out.ca),
            cmd_nphases_span = 4,
            extended_overlaps_check = extended_overlaps_check
        )

        self.comb += self.out.cs.eq(self.commands.cs)
        for bit in range(6):
            self.comb += self.out.ca[bit].eq(self.commands.ca[bit])

        # DQ ---------------------------------------------------------------------------------------
        dq_oe = Signal()
        self.comb += self.out.dq_oe.eq(delayed(self, dq_oe, cycles=1))

        for bit in range(self.databits):
            # output
            wrdata = [
                self.dfi.phases[i//2].wrdata[i%2 * self.databits + bit]
                for i in range(2*nphases)
            ]
            self.submodules += BitSlip(
                dw     = 2*nphases,
                cycles = bitslip_cycles,
                rst    = self.get_rst(bit//8, self._wdly_dq_bitslip_rst.re),
                slp    = self.get_inc(bit//8, self._wdly_dq_bitslip.re),
                i      = Cat(*wrdata),
                o      = self.out.dq_o[bit],
            )

            # input
            dq_i_bs = Signal(2*nphases)
            self.submodules += BitSlip(
                dw     = 2*nphases,
                cycles = bitslip_cycles,
                rst    = self.get_rst(bit//8, self._rdly_dq_bitslip_rst.re),
                slp    = self.get_inc(bit//8, self._rdly_dq_bitslip.re),
                i      = self.out.dq_i[bit],
                o      = dq_i_bs,
            )
            for i in range(2*nphases):
                self.comb += self.dfi.phases[i//2].rddata[i%2 * self.databits + bit].eq(dq_i_bs[i])

        # DQS --------------------------------------------------------------------------------------
        dqs_oe        = Signal()
        dqs_preamble  = Signal()
        dqs_postamble = Signal()
        dqs_pattern   = DQSPattern(
            preamble      = dqs_preamble,
            postamble     = dqs_postamble,
            wlevel_en     = self._wlevel_en.storage,
            wlevel_strobe = self._wlevel_strobe.re)
        self.submodules += dqs_pattern
        self.comb += [
            self.out.dqs_oe.eq(delayed(self, dqs_oe, cycles=1)),
        ]

        for byte in range(self.databits//8):
            # output
            self.submodules += BitSlip(
                dw     = 2*nphases,
                cycles = bitslip_cycles,
                rst    = self.get_rst(byte, self._wdly_dq_bitslip_rst.re),
                slp    = self.get_inc(byte, self._wdly_dq_bitslip.re),
                i      = dqs_pattern.o,
                o      = self.out.dqs_o[byte],
            )

        # DMI --------------------------------------------------------------------------------------
        # DMI signal is used for Data Mask or Data Bus Invertion depending on Mode Registers values.
        # With DM and DBI disabled, this signal is a Don't Care.
        # With DM enabled, masking is performed only when the command used is WRITE-MASKED.
        # We don't support DBI, DM support is configured statically with `masked_write`.
        for byte in range(self.databits//8):
            if isinstance(masked_write, Signal) or masked_write:
                self.comb += self.out.dmi_oe.eq(self.out.dq_oe)
                wrdata_mask = [
                    self.dfi.phases[i//2] .wrdata_mask[i%2 * self.databits//8 + byte]
                    for i in range(2*nphases)
                ]
                self.submodules += BitSlip(
                    dw     = 2*nphases,
                    cycles = bitslip_cycles,
                    rst    = self.get_rst(byte, self._wdly_dq_bitslip_rst.re),
                    slp    = self.get_inc(byte, self._wdly_dq_bitslip.re),
                    i      = Cat(*wrdata_mask),
                    o      = self.out.dmi_o[byte],
                )
            else:
                self.comb += self.out.dmi_o[byte].eq(0)
                self.comb += self.out.dmi_oe.eq(0)

        # Read Control Path ------------------------------------------------------------------------
        # Creates a delay line of read commands coming from the DFI interface. The output is used to
        # signal a valid read data to the DFI interface.
        #
        # The read data valid is asserted for 1 sys_clk cycle when the data is available on the DFI
        # interface, the latency is the sum of the OSERDESE2, CAS, ISERDESE2 and Bitslip latencies.
        rddata_en = TappedDelayLine(
            signal = reduce(or_, [dfi.phases[i].rddata_en for i in range(nphases)]),
            ntaps  = self.settings.read_latency
        )
        self.submodules += rddata_en

        self.comb += [
            phase.rddata_valid.eq(rddata_en.output | self._wlevel_en.storage)
            for phase in dfi.phases
        ]

        # Write Control Path -----------------------------------------------------------------------
        wrtap = cwl_sys_latency - 1
        assert wrtap >= 0

        # Create a delay line of write commands coming from the DFI interface. This taps are used to
        # control DQ/DQS tristates.
        wrdata_en = TappedDelayLine(
            signal = reduce(or_, [dfi.phases[i].wrdata_en for i in range(nphases)]),
            ntaps  = wrtap + 2
        )
        self.submodules += wrdata_en

        self.comb += dq_oe.eq(wrdata_en.taps[wrtap])
        # Always enabled in write leveling mode, else during transfers
        self.comb += dqs_oe.eq(self._wlevel_en.storage | (dqs_preamble | dq_oe | dqs_postamble))

        # Write DQS Postamble/Preamble Control Path ------------------------------------------------
        # Generates DQS Preamble 1 cycle before the first write and Postamble 1 cycle after the last
        # write. During writes, DQS tristate is configured as output for at least 3 sys_clk cycles:
        # 1 for Preamble, 1 for the Write and 1 for the Postamble.
        def wrdata_en_tap(i):  # allows to have wrtap == 0
            return wrdata_en.input if i == -1 else wrdata_en.taps[i]
        self.comb += dqs_preamble.eq( wrdata_en_tap(wrtap - 1)  & ~wrdata_en_tap(wrtap + 0))
        self.comb += dqs_postamble.eq(wrdata_en_tap(wrtap + 1)  & ~wrdata_en_tap(wrtap + 0))

    def get_rst(self, byte, rst):
        return (self._dly_sel.storage[byte] & rst) | self._rst.storage

    def get_inc(self, byte, inc):
        return self._dly_sel.storage[byte] & inc


class DoubleRateLPDDR4PHY(LPDDR4PHY):
    """LPDDR4PHY wrapper that performs one stage of serialization (16:8)

    Needed for targets that only have hardware serialization blocks up to 8:1.
    """
    def __init__(self, pads, *, ser_latency, des_latency, serdes_reset_cnt=0, **kwargs):
        super().__init__(pads,
            ser_latency = ser_latency + Latency(sys=Serializer.LATENCY),
            des_latency = des_latency + Latency(sys=Deserializer.LATENCY),
            **kwargs)

        self._out = self.out
        self.out = LPDDR4Output(nphases=self.nphases//2, databits=self.databits)

        def ser(i, o):
            assert len(o) == len(i)//2
            self.submodules += Serializer(
                clkdiv    = "sys",
                clk       = "sys2x",
                i_dw      = len(i),
                o_dw      = len(o),
                i         = i,
                o         = o,
                reset_cnt = serdes_reset_cnt,
            )

        def des(i, o):
            assert len(i) == len(o)//2
            self.submodules += Deserializer(
                clkdiv    = "sys",
                clk       = "sys2x",
                i_dw      = len(i),
                o_dw      = len(o),
                i         = i,
                o         = o,
                reset_cnt = serdes_reset_cnt,
            )

        # handle ser/des for both the lists (like dq) and just Signal (like cs)
        def apply(fn, i, o):
            if not isinstance(i, list):
                i, o = [i], [o]
            for i_n, o_n in zip(i, o):
                fn(i=i_n, o=o_n)

        for name in vars(self.out):
            old = getattr(self._out, name)
            new = getattr(self.out, name)
            if name.endswith("_oe"):  # OE signals need to be delayed
                self.comb += new.eq(delayed(self, old, cycles=Serializer.LATENCY))
            elif name.endswith("_i"):  # Deserialize inputs
                apply(des, o=old, i=new)
            else:  # All other signals are outputs
                apply(ser, i=old, o=new)



#
# This file is part of LiteDRAM.
#
# Copyright (c) 2021 Antmicro <www.antmicro.com>
# SPDX-License-Identifier: BSD-2-Clause

import math
from operator import or_
from functools import reduce
from collections import OrderedDict

from migen import *

from litex.soc.interconnect import stream
from litex.soc.interconnect.csr import AutoCSR
#
from litedram.common import TappedDelayLine
from litedram.phy.utils import edge, delayed
from litedram.phy.sim_utils import SimLogger, Timing, PulseTiming, log_level_getter

from litedram.phy.lpddr5.basephy import get_frange


CMD_INFO_LAYOUT = [
    ("we", 1),
    ("masked", 1),
    ("burst32", 1),
    ("bank", 4),
    ("row", 18),
    ("col", 6),
]

gtkw_dbg = {}

class LPDDR5Sim(Module, AutoCSR):
    """LPDDR5 DRAM simulation
    """
    def __init__(self, pads, *, ck_freq, wck_ck_ratio, log_level, logger_kwargs, check_timings=True):
        log_level = log_level_getter(log_level)

        self.clock_domains.cd_ck = ClockDomain(reset_less=True)
        self.clock_domains.cd_ck_n = ClockDomain(reset_less=True)
        self.comb += [
            self.cd_ck.clk.eq(pads.ck),
            self.cd_ck_n.clk.eq(~pads.ck),
        ]

        self.clock_domains.cd_wck = ClockDomain(reset_less=True)
        self.clock_domains.cd_wck_n = ClockDomain(reset_less=True)
        self.comb += [
            self.cd_wck.clk.eq(pads.wck),
            self.cd_wck_n.clk.eq(~pads.wck),
        ]

        # CommandsSim and DataSim communicate via this endpoint
        cmd_info = stream.Endpoint(CMD_INFO_LAYOUT)
        gtkw_dbg["cmd_info"] = cmd_info

        cmd = CommandsSim(pads, cmd_info, ck_freq=ck_freq, check_timings=check_timings,
            logger_kwargs=logger_kwargs, log_level=log_level)
        self.submodules.cmd = ClockDomainsRenamer("ck")(cmd)

        data = DataSim(pads, cmd_info,
            latency_ready = cmd.data_timer.ready_p,
            mode_regs     = cmd.mode_regs,
            ck_freq       = ck_freq,
            wck_ck_ratio  = wck_ck_ratio,
            logger_kwargs = logger_kwargs,
            log_level     = log_level
        )
        self.submodules.data = ClockDomainsRenamer("wck")(data)


def nested_case(mapping, *, on_leaf, variables, default=None, **kwargs):
    """Generate a nested Case from a mapping

    Parameters
    ----------
    mapping : dict or list
        This is a nested tree structure that maps given variable value to either another mapping
        or a final value.
    on_leaf : Signal or callable(*values, **kwargs)
        If it is a Signal, each tree leaf will assign it a value depending on current variables.
        If it is a callable, then it is called on each tree leaf with all concrete variable values
        passed as `values` and it should return Migen expressions.
    variables : list(tuple(str, Signal))
        List of variables (name + signal) for subsequent tree levels. First variable will be used
        in the outermost Case as `mapping[var]`, 2nd for `mapping[_][var]` and so on...
    default : callable(name, var)
        Optional callback that can add operations added as "default" case for each level.
    kwargs : dict
        User keyword args, passed for each call to `on_leaf`.
    """
    # call the recursive version with initial argument values
    return _nested_case(mapping,
        on_leaf      = on_leaf,
        variables    = variables,
        default      = default,
        values       = [],
        orig_mapping = mapping,
        **kwargs)

def index_recurive(indexable, indices):
    for i in indices:
        indexable = indexable[i]
    return indexable

def _nested_case(mapping, *, on_leaf, variables, default, values, orig_mapping, **kwargs):
    debug = False

    if len(variables) == 0:
        if debug:
            print(f'{" "* 2*len(values)}on_leaf({values})')
        if callable(on_leaf):
            return on_leaf(*values, **kwargs)
        elif isinstance(on_leaf, Signal):
            return on_leaf.eq(index_recurive(orig_mapping, values))
        else:
            raise TypeError(on_leaf)
    else:
        name, var = variables[0]
        cases = {}
        if isinstance(mapping, dict):
            keys = list(mapping.keys())
        else:
            keys = list(range(len(mapping)))
        if debug:
            print(f'{" "* 2*len(values)}Case({name}, <{keys}>')
        for key in keys:
            cases[key] = _nested_case(mapping[key],
                on_leaf      = on_leaf,
                variables    = variables[1:],
                default      = default,
                values       = values + [key],
                orig_mapping = orig_mapping,
                **kwargs)
        if default is not None:
            cases["default"] = default(name, var)
        if debug:
            print(f'{" "* 2*len(values)})')
        return Case(var, cases)


class ModeRegisters(Module, AutoCSR):
    MR_RESET = {
        3: 0b00000110,
        4: 0b10000000,
        17: 0b00111000,
        18: 0b10000000,  # DRAM starts with WCK:CK=2:1
        28: 0b00000100,
        41: 0b01100000,
    }
    FIELD_DEFS = dict(
        # (address, (highest bit, lowest bit)), bits are inclusive
        wl = (1, (7, 4)),
        rl = (2, (3, 0)),
        set_ab = (3, (5, 5)),
        bank_org = (3, (4, 3)),
        ckr = (18, (7, 7)),
    )

    def __init__(self, *, log_level, logger_kwargs):
        self.submodules.log = SimLogger(log_level=log_level("mr"), **logger_kwargs)

        self.mr = Array([
            Signal(8, reset=self.MR_RESET.get(addr, 0), name=f"mr{addr}")
            for addr in range(64)
        ])

        self.fields = fields = {}
        for name, (addr, (bit_hi, bit_lo)) in self.FIELD_DEFS.items():
            fields[name] = Signal(bit_hi - bit_lo + 1)
            self.comb += fields[name].eq(self.mr[addr][bit_lo:bit_hi+1])

        self.ckr = Signal(max=4+1)
        self.comb += Case(fields["ckr"], {0: self.ckr.eq(4), 1: self.ckr.eq(2)})

        self.set_ab = Signal(2)
        self.comb += self.set_ab.eq(fields["set_ab"])

        value_warning = lambda name, var: self.log.warn(f"Unexpected value for '{name}': %d", var)

        self.wl = Signal(max=16+1)
        self.comb += nested_case(
            # DVFSC disabled; mapping[wck:ck][Set A/B][OP[7:4]]
            mapping = {
                2: [
                    [4, 4, 6, 8, 8, 10],
                    [4, 6, 8, 10, 14, 16],
                ],
                4: [
                    [2, 2, 3, 4, 4, 5, 6, 6, 7, 8, 9, 9],
                    [2, 3, 4, 5, 7, 8, 9, 11, 12, 14, 15, 16],
                ],
            },
            on_leaf = self.wl,
            default = value_warning,
            variables = [
                ("wck:ck ratio", self.ckr),
                ("set A/B", self.set_ab),
                ("wl field", fields["wl"]),
            ],
        )

        self.rl = Signal(max=20+1)
        self.comb += nested_case(
            # Link ECC off, DVFSC disabled; mapping[wck:ck][Set][OP[3:0]]
            mapping = {
                2: [
                    [6, 8, 10, 12, 16, 18],
                    [6, 8, 10, 14, 16, 20],
                    [6, 8, 12, 14, 18, 20],
                ],
                4: [
                    [3, 4, 5, 6, 8, 9, 10, 12, 13, 15, 16, 17],
                    [3, 4, 5, 7, 8, 10, 11, 13, 14, 16, 17, 18],
                    [3, 4, 6, 7, 9, 10, 12, 14, 15, 17, 19, 20],
                ],
            },
            on_leaf = self.rl,
            default = value_warning,
            variables = [
                ("wck:ck ratio", self.ckr),
                ("set A/B", self.set_ab),
                ("rl field", fields["rl"]),
            ],
        )


class Sync(list):
    # Helper for combining comb and sync
    def __init__(self, arg):
        if not isinstance(arg, list):
            arg = [arg]
        super().__init__(arg)


class CommandsSim(Module, AutoCSR):
    def __init__(self, pads, cmd_info, *, ck_freq, check_timings, log_level, logger_kwargs):
        self.submodules.log = SimLogger(log_level=log_level("cmd"), **logger_kwargs)
        self.comb += self.log.info("Simulation start")

        self.cmd_info = cmd_info
        self.submodules.mode_regs = ModeRegisters(log_level=log_level, logger_kwargs=logger_kwargs)

        self.nbanks = 16
        self.active_banks = Array([Signal(name=f"bank{i}_active") for i in range(self.nbanks)])
        self.active_rows = Array([Signal(18, name=f"bank{i}_active_row") for i in range(self.nbanks)])

        # MPC operand
        self.mpc_op  = Signal(8)

        # ZQ calibration TODO: implement ZQC logic
        self.zqc_start = Signal()
        self.zqc_latch = Signal()

        # The captured command is delayed and the timer starts 1 cycle later:
        #       CK  --____----____----____----____----____--
        #       CS  __--------______________________________  (center-aligned to CK)
        #       CA  ____ppppNNNN____________________________  (center-aligned to CK DDR)
        # ca_p_pre  ______xxxxxxxx__________________________
        # ca_n_pre  __________xxxxxxxx______________________
        #   ca_p/n  ______________XXXXXXXX__________________  (phase-aligned to CK)
        #   timing  ______________________8-------7-------6-  (phase-aligned to CK)
        cs = Signal()
        cs_pre = Signal()
        ca_p_pre = Signal(7)
        ca_n_pre = Signal(7)
        self.ca_p = Signal(7)
        self.ca_n = Signal(7)
        self.sync.ck_n += ca_n_pre.eq(pads.ca)
        self.sync.ck += [
            cs_pre.eq(pads.cs),
            cs.eq(cs_pre),
            ca_p_pre.eq(pads.ca),
            self.ca_p.eq(ca_p_pre),
            self.ca_n.eq(ca_n_pre),
        ]

        self.handle_cmd = Signal()

        self.data_latency = Signal(max(len(self.mode_regs.wl), len(self.mode_regs.rl)))
        data_latency = Signal.like(self.data_latency)
        data_latency_reg = Signal.like(self.data_latency)
        self.submodules.data_timer = PulseTiming(data_latency)
        self.sync += If(self.data_timer.trigger,
            data_latency_reg.eq(self.data_latency)
        )
        self.comb += If(self.data_timer.trigger,
            data_latency.eq(self.data_latency),
        ).Else(
            data_latency.eq(data_latency_reg),
        ),

        cmds_enabled = Signal()
        allow_unhandled_cmd = Signal()
        cmd_handlers = OrderedDict(
            ACT = self.activate_handler(),
            PRE = self.precharge_handler(),
            REF = self.refresh_handler(),
            MRW = self.mrw_handler(),
            DATA = self.data_handler(),
            CAS = self.cas_handler(),
            MPC = self.mpc_handler(),
            # MRR
            # WFF/RFF?
            # RDC?
        )
        self.comb += [
            self.handle_cmd.eq(cmds_enabled & cs),
            If(self.handle_cmd & ~reduce(or_, cmd_handlers.values()),
                self.log.error("Unexpected command: CA_p=0b%07b CA_n=0b%07b", self.ca_p, self.ca_n)
            ),
            If(cs & ~cmds_enabled & ~allow_unhandled_cmd,
                self.log.warn("Received command when no commands should be sent: CA_p=0b%07b CA_n=0b%07b", self.ca_p, self.ca_n)
            ),
        ]

        check_timings = 1 if check_timings else 0
        ck = lambda t: math.ceil(t * ck_freq)
        ms, us, ns = 1e-3, 1e-6, 1e-9
        self.submodules.tinit0 = PulseTiming(ck(20*ms))  # (max) voltage-ramp at power-up; not applicable in the simulation
        self.submodules.tinit1 = PulseTiming(ck(200*us))  # (min) reset_n low time after voltage-ramp completion
        self.submodules.tinit2 = Timing(ck(10*ns))  # (min) CS low time before reset deassertion
        self.submodules.tinit3 = PulseTiming(ck(2*ms))  # (min) CS low time after reset deassertion
        self.submodules.tinit4 = PulseTiming(5)  # (min) stabilized CK before CS high; not really applicable in this simulation
        self.submodules.tinit5 = PulseTiming(ck(2*us))  # (min) idle time before first MRW/MRR cmmand
        self.submodules.tzqlat = PulseTiming(max(4, ck(30*ns)))  # (min) ZQCAL latch quiet time
        tpw_reset_ck = ck(100*ns) if check_timings else 4  # Avoids double reset during initialization at high frequecies
        self.submodules.tpw_reset = PulseTiming(tpw_reset_ck)  # (min) RESET_n low time for Reset initialization with stable power

        def with_progress(timing, string, *args, as_clocks=False):
            current, full = timing.progress()
            width = len(str(full)) if not isinstance(full, Signal) else len(str(2**len(full)))
            string += f" (%{width}d/%{width}d ck)"
            return (string, *args, current, full)

        self.comb += [
            self.tpw_reset.trigger.eq(~pads.reset_n),
            self.tinit2.valid.eq(~pads.cs),
            If(edge(self, pads.reset_n),
                If(~self.tinit2.ready & check_timings,
                    self.log.warn(*with_progress(self.tinit2, "tINIT2 violated: CS LOW for too short before deasserting RESET"))
                ),
                If(~self.tpw_reset.ready & check_timings,
                    self.log.warn(*with_progress(self.tpw_reset, "tPW_RESET violated: RESET_n held low for too short"))
                ),
            ),
            If(edge(self, pads.reset_n),
                self.log.info("RESET released"),
            ).Elif(edge(self, ~pads.reset_n),
                self.log.info("RESET asserted"),
            ),
        ]

        # We use an FSM that will be automatically reset to the RESET state when reset pad is asserted.
        # NOTE: for simulation purpose we assume that CK is always running because CommandsSim is clocked
        # from it, or else the states up to Tc would make no sense because the timings would not be counted
        class ResetFSM(FSM):
            def __init__(self, tpw_reset):
                self.tpw_reset = tpw_reset
                super().__init__(reset_state="RESET")

            def act(self, state, *statements):
                if state != "RESET":
                    statements = [
                        If(edge(self, self.tpw_reset.ready),
                            NextState("RESET"),
                        ).Else(*statements)
                    ]
                super().act(state, statements)

        self.submodules.fsm = fsm = ResetFSM(self.tpw_reset)
        fsm.act("RESET",
            self.tinit1.trigger.eq(1),
            If(edge(self, pads.reset_n),
                If(~self.tinit1.ready & check_timings,
                    self.log.warn(*with_progress(self.tinit1, "tINIT1 violated: RESET deasserted too fast"))
                ),
                NextState("WAIT-NOP")  # Tc
            ),
        )
        fsm.act("WAIT-NOP",
            self.tinit3.trigger.eq(1),
            If(cs & ~self.tinit3.ready & check_timings,
                self.log.warn(*with_progress(self.tinit3, "tINIT3 violated: CS high too fast after RESET deassertion"))
            ),
            self.tinit4.trigger.eq(1),
            If(cs & ~self.tinit4.ready & check_timings,
                self.log.warn(*with_progress(self.tinit4, "tINIT4 violated: CS high too fast after stable CK"))
            ),
            If(cs,  # NOP; TODO: DRAM probably only checks CS, then we'd better not check CA
                allow_unhandled_cmd.eq(1),
                self.tinit5.trigger.eq(1),
                If(~self.tinit4.ready & check_timings,
                    self.log.warn(*with_progress(self.tinit4, "tINIT4 violated: CS HIGH too fast while waiting for initial NOP"))
                ),
                If((self.ca_p != 0) | (self.ca_n != 0),
                    self.log.warn("Waiting for NOP but got CA_p=0b%07b CA_n=0b%07b", self.ca_p, self.ca_n)
                ),
                NextState("NO-CMDS"),  # Te
            )
        )
        fsm.act("NO-CMDS",
            self.tinit5.trigger.eq(1),
            If(cs & ~self.tinit5.ready & check_timings,
                self.log.warn(*with_progress(self.tinit5, "tINIT5 violated: command issued too fast after initial NOP"))
            ),
            If(self.tinit5.ready | ~check_timings,
                NextState("MODE-REGS")  # Tf
            )
        )
        fsm.act("MODE-REGS",
            cmds_enabled.eq(1),
            # If(self.handle_cmd & ~cmd_handlers["MRW"] & ~cmd_handlers["MRR"] & ~cmd_handlers["MPC"],  # TODO: MRR
            If(self.handle_cmd & ~cmd_handlers["MRW"] & ~cmd_handlers["MPC"],
                self.log.warn("Only MRW/MRR commands expected before ZQ Latch ..."),
                self.log.warn("... " + " ".join("{}=%d".format(cmd) for cmd in cmd_handlers.keys()), *cmd_handlers.values()),
            ),
            If(cmd_handlers["MPC"],
                If(~self.zqc_latch,
                    self.log.error("ZQC-LATCH expected, got MPC with op=0b%07b", self.mpc_op)
                ).Else(
                    NextState("ZQC-LATCH")  # Tg
                )
            ),
        )
        fsm.act("ZQC-LATCH",
            cmds_enabled.eq(1),
            self.tzqlat.trigger.eq(1),
            If(~self.tzqlat.ready & self.handle_cmd & check_timings,
                self.log.error(*with_progress(self.tzqlat, "tZQCAL violated: new command issued too fast: CA_p=0b%07b CA_n=0b%07b", self.ca_p, self.ca_n))
            ),
            If(self.tzqlat.ready | ~check_timings,
                NextState("NORMAL"),  # Th
            )
        )
        fsm.act("NORMAL",
            cmds_enabled.eq(1)
        )

        # Log state transitions
        fsm.finalize()
        prev_state = delayed(self, fsm.state)
        self.comb += If(prev_state != fsm.state,
            Case(prev_state, {
                state: Case(fsm.state, {
                    next_state: self.log.info(f"FSM: {state_name} -> {next_state_name}")
                    for next_state, next_state_name in fsm.decoding.items()
                })
                for state, state_name in fsm.decoding.items()
            })
        )

    def activate_handler(self):
        bank = Signal(max=self.nbanks)
        row1 = Signal(4)
        row2 = Signal(3)
        row3 = Signal(4)
        row4 = Signal(7)
        row = Signal(18)
        return self.cmd_two_step("ACTIVATE",
            cond1 = self.ca_p[:3] == 0b111,
            body1 = [
                NextValue(row1, self.ca_p[3:]),
                NextValue(row2, self.ca_n[4:]),
                NextValue(bank, self.ca_n[:4]),
            ],
            cond2 = self.ca_p[:3] == 0b011,
            body2 = [
                self.log.info("ACT: bank=%d row=%d", bank, row),
                row3.eq(self.ca_p[3:]),
                row4.eq(self.ca_n),
                row.eq(Cat(row4, row3, row2, row1)),
                NextValue(self.active_banks[bank], 1),
                NextValue(self.active_rows[bank], row),
                If(self.active_banks[bank],
                    self.log.error("ACT on already active bank: bank=%d row=%d", bank, row)
                ),
            ],
            wait_time = 8,  # tAAD
        )

    def precharge_handler(self):
        bank = Signal(max=self.nbanks)
        all_banks = Signal()
        return self.cmd_one_step("PRECHARGE",
            cond = self.ca_p[:7] == 0b1111000,
            body = [
                all_banks.eq(self.ca_n[6]),
                If(all_banks,
                    self.log.debug("PRE: all banks"),
                    bank.eq(2**len(bank) - 1),
                ).Else(
                    self.log.debug("PRE: bank = %d", bank),
                    bank.eq(self.ca_n[:4]),
                ),
                Sync(
                    If(all_banks,
                        *[self.active_banks[b].eq(0) for b in range(2**len(bank))]
                    ).Else(
                        self.active_banks[bank].eq(0),
                        If(~self.active_banks[bank],
                            self.log.warn("PRE on inactive bank: bank=%d", bank)
                        ),
                    ),
                )
            ],
        )

    def refresh_handler(self):
        # TODO: refresh tracking
        bank = Signal(max=self.nbanks)
        all_banks = Signal()
        return self.cmd_one_step("REFRESH",
            cond = self.ca_p[:7] == 0b0111000,
            body = [
                all_banks.eq(self.ca_n[6]),
                If(reduce(or_, self.active_banks),
                    self.log.error("Not all banks precharged during REFRESH")
                )
            ]
        )

    def mrw_handler(self):
        ma  = Signal(7)
        op  = Signal(8)
        return self.cmd_two_step("MRW",
            cond1 = self.ca_p[:7] == 0b1011000,
            body1 = [
                NextValue(ma, self.ca_n),
            ],
            cond2 = self.ca_p[:6] == 0b001000,
            body2 = [
                op.eq(Cat(self.ca_n, self.ca_p[6])),
                self.log.info("MRW: MR[%d] = 0x%02x", ma, op),
                NextValue(self.mode_regs.mr[ma], op),
            ]
        )

    def mpc_handler(self):
        op  = Signal(8)
        return self.cmd_one_step("MRW",
            cond = self.ca_p[:6] == 0b110000,
            body = [
                op.eq(Cat(self.ca_n, self.ca_p[6])),
                self.mpc_op.eq(op),
                self.zqc_start.eq(op == 0b10000101),
                self.zqc_latch.eq(op == 0b10000110),
                If(~(self.zqc_start | self.zqc_latch),
                    self.log.warn("Unsupported MPC command: op=0b%08b", op),
                ),
            ],
        )

    def cas_handler(self):
        ws_wr = Signal()
        ws_rd = Signal()
        ws_fs = Signal()
        dc    = Signal(4)
        wrx   = Signal()
        wxsa  = Signal()
        wxsb  = Signal()
        return self.cmd_one_step("CAS",
            cond = self.ca_p[:4] == 0b1100,
            body = [
                # TODO: implement WCK sync
                ws_wr.eq(self.ca_p[4]),
                ws_rd.eq(self.ca_p[5]),
                ws_fs.eq(self.ca_p[6]),
                dc.eq(self.ca_n[:4]),
                wrx.eq(self.ca_n[4]),
                wxsa.eq(self.ca_n[5]),
                wxsb.eq(self.ca_n[6]),
                If(sum([ws_wr, ws_rd, ws_fs]) > 1,
                    self.log.error("More than one WCK Sync bit in CAS command")
                ),
                If(reduce(or_, [dc, wrx, wxsa, wxsb]),
                    self.log.warn("Unsupported CAS function requested")
                ),
            ]
        )

    def data_handler(self):
        data_cmds = {
            "MASKED-WRITE": self.ca_p[:3] == 0b010,
            "WRITE":        self.ca_p[:3] == 0b110,
            "WRITE32":      self.ca_p[:4] == 0b0100,
            "READ":         self.ca_p[:3] == 0b001,
            "READ32":       self.ca_p[:3] == 0b101,
        }

        bank           = Signal(max=self.nbanks)
        row            = Signal(18)
        col            = Signal(6)
        auto_precharge = Signal()

        return self.cmd_one_step("DATA",
            cond = reduce(or_, data_cmds.values()),
            body = [
                bank.eq(self.ca_n[:4]),
                row.eq(self.active_rows[bank]),
                col.eq(Cat(self.ca_p[3], self.ca_n[4:6], self.ca_p[4:7])),
                auto_precharge.eq(self.ca_n[6]),
                # push to DataSim
                self.cmd_info.we.eq(data_cmds["MASKED-WRITE"] | data_cmds["WRITE"] | data_cmds["WRITE32"]),
                self.cmd_info.masked.eq(data_cmds["MASKED-WRITE"]),
                self.cmd_info.burst32.eq(data_cmds["WRITE32"] | data_cmds["READ32"]),
                self.cmd_info.bank.eq(bank),
                self.cmd_info.row.eq(row),
                self.cmd_info.col.eq(col),
                self.cmd_info.valid.eq(1),
                # data latency
                If(self.cmd_info.we,
                    self.data_latency.eq(self.mode_regs.wl - 1),
                    If(self.mode_regs.wl < 1,
                        self.log.error("WL < 2 is not supported")
                    ),
                ).Else(
                    # FIXME: Currently we need to subtract 1 cycle here and delay additionally in BurstHalf.
                    # We need to check if that's a limitation of PHY not being able to level DRAM only in simulation
                    # or if there's a need to increase read_latency by 1 cycle (or increase bitslip cycles).
                    self.data_latency.eq(self.mode_regs.rl - 2),
                    If(self.mode_regs.rl < 2,
                        self.log.error("RL < 2 is not supported")
                    ),
                ),
                self.data_timer.trigger.eq(1),
                # command info
                *[If(cond, self.log.info(f"{name}: bank=%d row=%d col=%d", bank, row, col))
                    for name, cond in data_cmds.items()],
                # auto precharge
                If(auto_precharge,
                    self.log.info("AUTO-PRECHARGE: bank=%d row=%d", bank, row),
                ),
                Sync(If(auto_precharge,
                    self.active_banks[bank].eq(0),
                )),
                # sanity checks
                If(~self.active_banks[bank],
                    self.log.error("CAS command on inactive bank: bank=%d row=%d col=%d", bank, row, col)
                ),
                If(self.cmd_info.masked & ~((self.mode_regs.fields["bank_org"] == 0b00) | (self.mode_regs.fields["bank_org"] == 0b10)),
                    self.log.error("READ32/WRITE32 are valid in BG/16B mode only")
                ),
            ],
        )

    def cmd_one_step(self, name, cond, body):
        matched = Signal()
        comb = list(filter(lambda i: not isinstance(i, Sync), body))
        sync = list(filter(lambda i: isinstance(i, Sync), body))
        self.comb += If(self.handle_cmd & cond,
            self.log.debug(name),
            matched.eq(1),
            *comb
        )
        if len(sync) > 0:
            self.sync += If(self.handle_cmd & cond,
                *sync
            )
        return matched

    def cmd_two_step(self, name, cond1, body1, cond2, body2, wait_time=None):
        state1, state2 = f"{name}-1", f"{name}-2"
        matched = Signal()

        timeout = 0
        trigger_timer = []
        if wait_time is not None:
            next_cmd_timer = PulseTiming(wait_time)
            self.submodules += next_cmd_timer
            timeout = next_cmd_timer.ready
            trigger_timer = [
                next_cmd_timer.trigger.eq(1),
            ]

        fsm = FSM()
        fsm.act(state1,
            If(self.handle_cmd & cond1,
                self.log.debug(state1),
                matched.eq(1),
                *body1,
                *trigger_timer,
                NextState(state2)
            )
        )
        fsm.act(state2,
            If(timeout,
                self.log.warn(f"Timeout while waiting for {state2}"),
                NextState(state1)
            ).Elif(self.handle_cmd,
                If(cond2,
                    self.log.debug(state2),
                    matched.eq(1),
                    *body2
                ).Else(
                    self.log.error(f"Waiting for {state2} but got unexpected CA_p=0b%07b CA_n=0b%07b", self.ca_p, self.ca_n)
                ),
                NextState(state1)  # always back to first
            )
        )
        self.submodules += fsm

        return matched


class DataSim(Module, AutoCSR):
    def __init__(self, pads, cmd_info, latency_ready, mode_regs, ck_freq, wck_ck_ratio, *, log_level, logger_kwargs, nrows=32768, ncols=1024, nbanks=16):
        self.submodules.log = SimLogger(log_level=log_level("data"), **logger_kwargs)

        # CommandsSim produces the data required for handling a data command via cmd_info endpoint.
        # Using stream.ClockDomainCrossing introduces too much latency, so we do a simplistic CDC
        # and store the information in a FIFO, so that it is possible to pipeline data commands.
        self.submodules.cmd_buf = stream.PipeValid(CMD_INFO_LAYOUT)
        gtkw_dbg["cmd_buf"] = self.cmd_buf

        # The CommandsSim data handling generated command is clocked under the system clock domain.
        # Instead, the DataSim is clocked in the wck clock domain.
        # Given that, with a dynamic wck regime, wck needs to get through a preamble phase where it
        # is not active for a certain amount of clock cycles, the command buffer pipeline may not
        # be able to see changes in the data commands generated by CommandSim.
        #
        # The following logic verifies whether a delay needs to be added and adds it accordingly to
        # the data command, so that the pipe stream can intercept the various commands from the ck
        # domain correctly.
        twck = 1 / (ck_freq * wck_ck_ratio)
        frange = get_frange(twck, wck_ck_ratio).for_set(wl_set="A", rl_set=0)
        taps = max(1, max(frange.t_wckenl_wr, frange.t_wckenl_rd) + frange.t_wckpre_static) - 1

        def delay_cmd_info(signals):
            for signal in signals:
                tapped_delay = ClockDomainsRenamer("ck")(TappedDelayLine(getattr(cmd_info, signal), ntaps=taps))
                setattr(self.submodules, f"{signal}_tap", tapped_delay)

                self.comb += [
                    getattr(self.cmd_buf.sink, signal).eq(reduce(or_, getattr(self, f"{signal}_tap").taps[0:taps])),
                ]

        if taps > 0:
            delay_cmd_info([tup[0] for tup in CMD_INFO_LAYOUT] + ["valid", "ready", "first", "last"])

        else:
            self.comb += [
                cmd_info.connect(self.cmd_buf.sink, omit={"ready", "valid"}),
                # to latch a command only once we use an edge here, which we can do as there is no way
                # for 2 valid commands cycle-by-cycle
                self.cmd_buf.sink.valid.eq(edge(self, cmd_info.valid)),
                # if for some reason buffer hasn't been cleared, then we have an error
                If(self.cmd_buf.sink.valid & ~self.cmd_buf.sink.ready,
                    self.log.error("Simulator internal error: CMD-to-DATA overflow")
                ),
            ]

        self.comb += [
            cmd_info.ready.eq(1),
        ]

        cmd = self.cmd_buf.source

        # DRAM Memory storage
        mems = [Memory(len(pads.dq), depth=nrows * ncols) for _ in range(nbanks)]
        self.specials += mems

        # Combine SDR data from each burst handler into DDR value
        dq_i_p = Signal.like(pads.dq_i)
        dq_i_n = Signal.like(pads.dq_i)
        self.comb += [
            If(pads.wck,
                pads.dq_i.eq(dq_i_n),
            ).Else(
                pads.dq_i.eq(dq_i_p),
            )
        ]

        Burst = lambda wck_cd, dq_i_x: BurstHalf(wck_cd=wck_cd, pads=pads, dq_i=dq_i_x, cmd=cmd,
            mems=mems, nrows=nrows, ncols=ncols, log_level=log_level, logger_kwargs=logger_kwargs)
        self.submodules.burst_p = ClockDomainsRenamer("wck")(Burst("wck", dq_i_p))
        self.submodules.burst_n = ClockDomainsRenamer("wck_n")(Burst("wck_n", dq_i_n))

        def delay(sig, cycles):
            if cycles == 0:
                return sig
            return delayed(self, sig, cycles=cycles)

        # After the WL signal arives we require the data to arrive some time later and then we start
        # reading it. This would be adjustable on hardware, but in simulation we rather must set this
        # so that it matches the delay that PHY introduces.
        wr_start = TappedDelayLine(ntaps=2)
        rd_start = TappedDelayLine(ntaps=2)
        self.submodules += wr_start, rd_start

        def delayed_cases(signal, delay_line, ckr_to_delay):
            cases = {}
            for ckr, delay in ckr_to_delay.items():
                cases[ckr] = signal.eq(delay_line.input if delay == 0 else delay_line.taps[delay - 1])
            return Case(mode_regs.ckr, cases)

        self.comb += [
            wr_start.input.eq(cmd.valid & cmd.we & latency_ready),
            rd_start.input.eq(cmd.valid & ~cmd.we & latency_ready),
            delayed_cases(self.burst_p.enable_wr, wr_start, {2: 0, 4: 0}),
            delayed_cases(self.burst_n.enable_wr, wr_start, {2: 1, 4: 1}),
            delayed_cases(self.burst_p.enable_rd, rd_start, {2: 0, 4: 0}),
            delayed_cases(self.burst_n.enable_rd, rd_start, {2: 1, 4: 1}),
            cmd.ready.eq(self.burst_p.ready),
        ]


class BurstHalf(Module):
    def __init__(self, *, pads, dq_i, cmd, mems, wck_cd, nrows, ncols, log_level, logger_kwargs):
        self.submodules.log = SimLogger(log_level=log_level("burst"), **logger_kwargs)

        self.enable_wr = Signal()
        self.enable_rd = Signal()
        self.ready     = Signal()

        # Register the command
        cmd_d = stream.Endpoint(CMD_INFO_LAYOUT)
        self.sync += cmd_d.eq(cmd)

        # Memory interface
        burst_start = {"wck": 0, "wck_n": 1}[wck_cd]
        mem_addr    = Signal(max=nrows * ncols)
        current_col = Signal(max=ncols)
        col_num     = Signal.like(current_col)
        burst_beat  = Signal.like(current_col, reset=burst_start)

        ports = [
            mem.get_port(write_capable=True, we_granularity=8, async_read=True, clock_domain=wck_cd)
            for mem in mems
        ]
        self.specials += ports
        ports = Array(ports)
        we_all = Signal.like(ports[0].we, reset=2**len(ports[0].we) - 1)

        # Burst control
        burst_length = Signal.like(burst_beat)
        self.comb += [
            self.ready.eq(burst_beat[1:] == burst_length[1:]),
            If(cmd_d.burst32,
                burst_length.eq(32 - 1)
            ).Else(
                burst_length.eq(16 - 1)
            ),
        ]

        self.comb += [
            col_num[4:].eq(cmd_d.col),
            current_col.eq(col_num + burst_beat),
            mem_addr.eq(cmd_d.row * ncols + current_col),
            ports[cmd_d.bank].adr.eq(mem_addr),
        ]

        self.submodules.fsm = fsm = FSM()
        fsm.act("IDLE",
            If(self.enable_wr | self.enable_rd,
                NextValue(burst_beat, burst_start),
                If(self.enable_wr,
                    NextState("BURST-WRITE"),
                ).Else(
                    NextState("BURST-READ"),
                ),
            )
        )
        fsm.act("BURST-WRITE",
            If(cmd_d.masked,
                ports[cmd_d.bank].we.eq(delayed(self, ~pads.dmi)),  # DMI HIGH masks a byte
            ).Else(
                ports[cmd_d.bank].we.eq(delayed(self, we_all)),
            ),
            ports[cmd_d.bank].dat_w.eq(delayed(self, pads.dq)),
            self.log.debug("WRITE[%d]: bank=%d, row=%d, col=%d, dq=0x%04x dm=0x%02b",
                burst_beat, cmd_d.bank, cmd_d.row, current_col, delayed(self, pads.dq), pads.dmi,
                once=False
            ),
            If(self.ready,
                NextValue(burst_beat, burst_start),
                If(~self.enable_wr,
                    NextState("IDLE"),
                ),
            ).Else(
                NextValue(burst_beat, burst_beat + 2),
            ),
        )
        fsm.act("BURST-READ",
            ports[cmd_d.bank].we.eq(0),
            dq_i.eq(ports[cmd_d.bank].dat_r),
            self.log.debug("READ[%d]: bank=%d, row=%d, col=%d, dq=0x%04x",
                burst_beat, cmd_d.bank, cmd_d.row, current_col, pads.dq,
                once=False
            ),
            If(self.ready,
                NextValue(burst_beat, burst_start),
                If(~self.enable_rd,
                    NextState("IDLE"),
                ),
            ).Else(
                NextValue(burst_beat, burst_beat + 2),
            ),
        )
